"""Builtin "bool" type.

There are several classes that are used to represent bool expressions:
- BoolLiteral: boolean literal (True or False).
- BoolVar: boolean variable. We use a score to store a boolean.
  When a score == 1, it's True; when score == 0, it's False.
- NotBoolVar: inverted (`not var`) form of `BoolVar`.
- CompareBase: abstract class that should be returned by
  `AcaciaExpr.compare` to implement comparison operators. 
- WildBool: stores any boolean expression. It consists of two parts:
  the command dependencies and /execute subcommands. The expression
  is True if and only if the subcommands executed successfully.
  It also stores "ranges" of a specific scoreboard slot (this is used
  as an optimization for "and" expressions).
- NotWildBool: inverted (`not expr`) form of `WildBool`.

Q: WHERE ARE `and` AND `or` EXPRESSION STORED???
A: `and` expressions are created using `new_and_group`. Normally a
`WildBool` is returned but some optimizations may apply. `or`
expressions are all converted into other types of expressions. This
is done by `new_or_expression`.
"""

__all__ = [
    # Type
    'BoolType', 'BoolDataType',
    # Expressions
    'BoolLiteral', 'BoolVar', 'NotBoolVar', 'CompareBase', 'WildBool',
    'NotWildBool', 'SupportsAsExecute',
    # Factory functions
    'new_and_group', 'new_or_expression',
    # Utils
    'to_BoolVar'
]

from typing import List, Tuple, Dict, Optional, TYPE_CHECKING
from abc import ABCMeta, abstractmethod

from .types import Type
from acaciamc.ast import Operator, COMPOP_INVERT
from acaciamc.error import *
from acaciamc.constants import INT_MAX, INT_MIN
from acaciamc.mccmdgen.expr import *
from acaciamc.mccmdgen.datatype import (
    DefaultDataType, Storable, SupportsEntityField
)
from acaciamc.mccmdgen.ctexpr import CTDataType
from acaciamc.mccmdgen.utils import InvalidOpError
import acaciamc.mccmdgen.cmds as cmds

if TYPE_CHECKING:
    from acaciamc.mccmdgen.cmds import _ExecuteSubcmd
    from acaciamc.compiler import Compiler
    from acaciamc.mccmdgen.ctexpr import CTObj

def _bool_compare(self: AcaciaExpr, op: Operator, other: AcaciaExpr, compiler):
    """Wildcard `compare` method to compare a bool with a bool."""
    if not (op is Operator.equal_to or op is Operator.unequal_to):
        raise InvalidOpError
    dep = []
    invert = op is Operator.unequal_to
    if isinstance(other, BoolVar):
        slot2 = other.slot
    elif isinstance(other, NotBoolVar):
        slot2 = other.slot
        invert = not invert
    elif isinstance(other, (CompareBase, WildBool)):
        _dep, other_var = to_BoolVar(other, compiler)
        dep.extend(_dep)
        slot2 = other_var.slot
    else:
        raise InvalidOpError
    _dep, self_var = to_BoolVar(self, compiler)
    dep.extend(_dep)
    return ScbEqualCompare(self_var.slot, slot2, invert, dep)

def _export_bool(subcmds: List["_ExecuteSubcmd"], var: "BoolVar",
                 compiler: "Compiler", invert=False) -> CMDLIST_T:
    res = []
    FALSE = int(invert)
    TRUE = 1 - FALSE
    # decide if temporary value is needed
    need_tmp = False
    for subcmd in subcmds:
        if subcmd.scb_did_read(var.slot):
            need_tmp = True
            break
    if need_tmp:
        tmp = compiler.allocate_tmp()
    else:
        tmp = var.slot
    # set target to False first
    res.append(cmds.ScbSetConst(tmp, FALSE))
    # check condition and change it to True
    res.append(cmds.Execute(subcmds, cmds.ScbSetConst(tmp, TRUE)))
    if need_tmp:
        res.append(cmds.ScbOperation(cmds.ScbOp.ASSIGN, var.slot, tmp))
    return res

def _bool_rawtextify(self: AcaciaExpr, compiler: "Compiler"):
    """A generic `rawtextify` implementation for bool."""
    dependencies, var = to_BoolVar(self, compiler)
    return [cmds.RawtextScore(var.slot)], dependencies

class BoolDataType(DefaultDataType, Storable, SupportsEntityField):
    name = "bool"

    def new_var(self, compiler) -> "BoolVar":
        return BoolVar.new(compiler)

    def new_entity_field(self, compiler):
        return {"scoreboard": compiler.add_scoreboard()}

    def new_var_as_field(self, entity, scoreboard: str) -> "BoolVar":
        return BoolVar(cmds.ScbSlot(entity.to_str(), scoreboard))

ctdt_bool = CTDataType("bool")

class BoolType(Type):
    def datatype_hook(self):
        return BoolDataType()

    def cdatatype_hook(self):
        return ctdt_bool

class SupportsAsExecute(AcaciaExpr, metaclass=ABCMeta):
    """See `as_execute` method. Should only be used by booleans."""
    @abstractmethod
    def as_execute(self, compiler) -> Tuple[CMDLIST_T, List["_ExecuteSubcmd"]]:
        """
        Convert this expression to some dependency commands and
        subcommands of /execute. Value of this expression is considered
        True if and only if the subcommands executed successfully.
        """
        pass

class BoolLiteral(ConstExprCombined):
    """Literal boolean."""
    cdata_type = ctdt_bool

    def __init__(self, value: bool):
        super().__init__(BoolDataType())
        self.value = value

    def export(self, var: "BoolVar", compiler):
        return [cmds.ScbSetConst(var.slot, int(self.value))]

    def cstringify(self) -> str:
        return "true" if self.value else "false"

    def rawtextify(self, compiler):
        return [cmds.RawtextText("1" if self.value else "0")], []

    def chash(self):
        return self.value

    def compare(self, op: Operator, other: AcaciaExpr, compiler):
        if not (op is Operator.equal_to or op is Operator.unequal_to):
            raise InvalidOpError
        # For a boolean expression b, b == True is equivalent to "b"
        # b == False is equivalent to "not b"; != is the opposite
        true = self.value == (op is Operator.equal_to)
        if self.data_type.is_type_of(other):
            return other if true else other.unarynot(compiler)

    def ccompare(self, op: Operator, other: "CTObj"):
        if not (op is Operator.equal_to or op is Operator.unequal_to):
            raise InvalidOpError
        true = self.value == (op is Operator.equal_to)
        if isinstance(other, BoolLiteral):
            return other.value == true
        raise InvalidOpError

    def __str__(self):
        """Return 1 if True, 0 if False."""
        return str(int(self.value))

    # Unary operator
    def cunarynot(self):
        return BoolLiteral(not self.value)

class BoolVar(VarValue, SupportsAsExecute):
    """Boolean stored as a score on scoreboard."""
    def __init__(self, slot: cmds.ScbSlot):
        super().__init__(BoolDataType())
        self.slot = slot

    @classmethod
    def new(cls, compiler: "Compiler", tmp=False):
        alloc = compiler.allocate_tmp if tmp else compiler.allocate
        return cls(alloc())

    def as_execute(self, compiler):
        return [], [cmds.ExecuteScoreMatch(self.slot, '1')]

    def __str__(self):
        return self.slot.to_str()

    def export(self, var: "BoolVar", compiler):
        return [cmds.ScbOperation(cmds.ScbOp.ASSIGN, var.slot, self.slot)]

    def compare(self, op, other, compiler):
        if not (op is Operator.equal_to or op is Operator.unequal_to):
            raise InvalidOpError
        if not isinstance(other, BoolVar):
            raise InvalidOpError
        invert = op is Operator.unequal_to
        return ScbEqualCompare(self.slot, other.slot, invert)

    def swap(self, other: "BoolVar", compiler):
        return [cmds.ScbOperation(cmds.ScbOp.SWAP, self.slot, other.slot)]

    rawtextify = _bool_rawtextify

    # Unary operator
    def unarynot(self, compiler):
        return NotBoolVar(self.slot)

class NotBoolVar(AcaciaExpr):
    def __init__(self, slot: cmds.ScbSlot):
        super().__init__(BoolDataType())
        self.slot = slot

    def __str__(self):
        return self.slot.to_str()

    def as_execute(self, compiler):
        return [], [cmds.ExecuteScoreMatch(self.slot, '0')]

    def export(self, var: BoolVar, compiler):
        if var.slot == self.slot:
            # negation of self
            return [
                cmds.ScbAddConst(var.slot, 1),
                cmds.Execute([cmds.ExecuteScoreMatch(var.slot, "2")],
                             cmds.ScbSetConst(var.slot, 0))
            ]
        else:
            return [
                cmds.ScbSetConst(var.slot, 0),
                cmds.Execute([cmds.ExecuteScoreMatch(self.slot, "0")],
                             cmds.ScbSetConst(var.slot, 1))
            ]

    def compare(self, op, other, compiler):
        if not (op is Operator.equal_to or op is Operator.unequal_to):
            raise InvalidOpError
        if not isinstance(other, (BoolVar, NotBoolVar)):
            raise InvalidOpError
        invert = (op is Operator.equal_to) == (isinstance(other, BoolVar))
        return ScbEqualCompare(self.slot, other.slot, invert)

    rawtextify = _bool_rawtextify

    # Unary operator
    def unarynot(self, compiler):
        res = BoolVar(self.slot)
        res.is_temporary = True
        return res

class CompareBase(SupportsAsExecute, metaclass=ABCMeta):
    def __init__(self):
        super().__init__(BoolDataType())

    def export(self, var: BoolVar, compiler):
        res, subcmds = self.as_execute(compiler)
        res.extend(_export_bool(subcmds, var, compiler))
        return res

    compare = _bool_compare
    rawtextify = _bool_rawtextify

    @abstractmethod
    def unarynot(self, compiler):
        # Boolean expressions should implement "not".
        pass

class ScbMatchesCompare(CompareBase):
    def __init__(
        self, dependencies: CMDLIST_T, slot: cmds.ScbSlot,
        operator: Operator, literal: int
    ):
        super().__init__()
        self.dependencies = dependencies
        self.slot = slot
        self.operator = operator
        self.literal = literal

    def as_execute(self, compiler) -> Tuple[CMDLIST_T, List["_ExecuteSubcmd"]]:
        res = []
        match_range = {
            Operator.greater: '%d..' % (self.literal + 1),
            Operator.greater_equal: '%d..' % self.literal,
            Operator.less: '..%d' % (self.literal - 1),
            Operator.less_equal: '..%d' % self.literal,
            Operator.equal_to: str(self.literal),
            Operator.unequal_to: '!%d' % self.literal
        }[self.operator]
        res.append(cmds.ExecuteScoreMatch(self.slot, match_range))
        return self.dependencies.copy(), res

    # Unary operator
    def unarynot(self, compiler):
        return ScbMatchesCompare(
            self.dependencies.copy(), self.slot, COMPOP_INVERT[self.operator],
            self.literal
        )

class ScbEqualCompare(CompareBase):
    def __init__(self, left: cmds.ScbSlot, right: cmds.ScbSlot,
                 invert=False, dependencies: Optional[CMDLIST_T] = None):
        super().__init__()
        self.left = left
        self.right = right
        self.invert = invert
        if dependencies is None:
            self.dependencies = []
        else:
            self.dependencies = dependencies

    def as_execute(self, compiler) -> Tuple[CMDLIST_T, List["_ExecuteSubcmd"]]:
        return (
            self.dependencies.copy(),
            [cmds.ExecuteScoreComp(
                self.left, self.right, cmds.ScbCompareOp.EQ, self.invert
            )]
        )

    def unarynot(self, compiler):
        return ScbEqualCompare(
            self.left, self.right, not self.invert,
            self.dependencies
        )

MCRange = Tuple[cmds.ScbSlot, int, int]

def _ranges2subcmds(ranges: List[MCRange]) -> List["_ExecuteSubcmd"]:
    res = []
    for slot, min_, max_ in ranges:
        if min_ == INT_MIN and max_ == INT_MAX:
            continue
        if min_ == max_:
            int_range = str(min_)
        else:
            min_str = '' if min_ == INT_MIN else str(min_)
            max_str = '' if max_ == INT_MAX else str(max_)
            int_range = '%s..%s' % (min_str, max_str)
        res.append(cmds.ExecuteScoreMatch(slot, int_range))
    return res

class WildBool(SupportsAsExecute):
    def __init__(
        self, subcmds: List["_ExecuteSubcmd"],
        dependencies: CMDLIST_T,
        ranges: Optional[List[MCRange]] = None
    ):
        super().__init__(BoolDataType())
        self.subcmds = subcmds
        self.dependencies = dependencies
        if ranges is None:
            ranges = []
        self.ranges = ranges

    def export(self, var: BoolVar, compiler):
        res, subcmds = self.as_execute(compiler)
        res.extend(_export_bool(subcmds, var, compiler))
        return res

    def as_execute(self, compiler):
        subcmds = self.subcmds + _ranges2subcmds(self.ranges)
        return self.dependencies.copy(), subcmds

    compare = _bool_compare
    rawtextify = _bool_rawtextify

    # Unary operator
    def unarynot(self, compiler):
        return NotWildBool(
            self.subcmds.copy(), self.dependencies.copy(),
            self.ranges.copy()
        )

class NotWildBool(AcaciaExpr):
    def __init__(
        self, subcmds: List["_ExecuteSubcmd"],
        dependencies: CMDLIST_T,
        ranges: Optional[List[MCRange]] = None
    ):
        super().__init__(BoolDataType())
        self.subcmds = subcmds
        self.dependencies = dependencies
        if ranges is None:
            ranges = []
        self.ranges = ranges

    def export(self, var: BoolVar, compiler):
        subcmds = self.subcmds + _ranges2subcmds(self.ranges)
        res = self.dependencies.copy()
        res.extend(_export_bool(subcmds, var, compiler, invert=True))
        return res

    compare = _bool_compare
    rawtextify = _bool_rawtextify

    # Unary operator
    def unarynot(self, compiler):
        return WildBool(
            self.subcmds.copy(), self.dependencies.copy(),
            self.ranges.copy()
        )

def new_and_group(operands: List[AcaciaExpr], compiler) -> AcaciaExpr:
    """Creates an "and" expression."""
    assert operands
    literals = [operand for operand in operands
                if isinstance(operand, BoolLiteral)]
    new_operands = operands.copy()
    for literal in literals:
        if literal.value:
            # throw away operands that are always true
            new_operands.remove(literal)
        else:
            # when any operand is literal false, the result is false
            return BoolLiteral(False)
    # if there is no operand left, return true
    if not new_operands:
        return BoolLiteral(True)
    # if there is only 1 operand left, return that
    if len(new_operands) == 1:
        return new_operands[0]
    # fallback
    res_subcmds: List[_ExecuteSubcmd] = []
    res_commands: CMDLIST_T = []
    ranges: Dict[cmds.ScbSlot, List[int]] = {}
    for operand in new_operands:
        if isinstance(operand, WildBool):
            res_subcmds.extend(operand.subcmds)
            res_commands.extend(operand.dependencies)
            for slot, min_, max_ in operand.ranges:
                rng = ranges.setdefault(slot, [INT_MIN, INT_MAX])
                rng[0] = max(min_, rng[0])
                rng[1] = min(max_, rng[1])
            operand = None
        elif isinstance(operand, NotWildBool):
            # `a and not (b and c)` needs a tmp var:
            # tmp = `not (b and c)`, self = `a and tmp`
            tmp = BoolVar.new(tmp=True, compiler=compiler)
            res_commands.extend(operand.export(tmp, compiler))
            operand = tmp
        elif (isinstance(operand, ScbMatchesCompare)
              and operand.operator is not Operator.unequal_to):
            # Optimization: When the same expr is compared to two or more
            # `IntLiteral`s, execute score matches subcommands can be merged.
            # e.g. `1 <= x <= 5 and x <= 3` can be merged into a single
            # /execute if score ... matches 1..3
            rng = ranges.setdefault(operand.slot, [INT_MIN, INT_MAX])
            res_commands.extend(operand.dependencies)
            literal = operand.literal
            if operand.operator is Operator.equal_to:
                rng[0] = max(literal, rng[0])
                rng[1] = min(literal, rng[1])
            elif operand.operator is Operator.greater:
                rng[0] = max(literal + 1, rng[0])
            elif operand.operator is Operator.greater_equal:
                rng[0] = max(literal, rng[0])
            elif operand.operator is Operator.less:
                rng[1] = min(literal - 1, rng[1])
            elif operand.operator is Operator.less_equal:
                rng[1] = min(literal, rng[1])
            if rng[0] > rng[1]:  # must be False
                return BoolLiteral(False)
            operand = None
        if operand is not None:
            assert isinstance(operand, (BoolVar, NotBoolVar, CompareBase))
            commands, subcmds = operand.as_execute(compiler)
            res_commands.extend(commands)
            res_subcmds.extend(subcmds)
    res_ranges = [(slot, *rng) for slot, rng in ranges.items()
                  if rng[0] != INT_MIN or rng[1] != INT_MAX]
    # if there is no subcommands generated, then just return True since
    # no restrictions is added.
    if not res_subcmds and not res_ranges:
        return BoolLiteral(True)
    # Final fallback
    return WildBool(res_subcmds, res_commands, res_ranges)

def new_or_expression(operands: List[AcaciaExpr], compiler) -> AcaciaExpr:
    """Create an "or" expression."""
    # invert the operands (`a`, `b`, `c` -> `not a`, `not b`, `not c`)
    inverted_operands = [operand.unarynot(compiler) for operand in operands]
    # connect them with `and` (-> `not a and not b and not c`)
    res = new_and_group(operands=inverted_operands, compiler=compiler)
    # invert result (-> `not (not a and not b and not c)`)
    return res.unarynot(compiler)

# Utils
def to_BoolVar(expr: AcaciaExpr, compiler, tmp=True) \
        -> Tuple[CMDLIST_T, BoolVar]:
    """Convert any boolean expression to a `BoolVar` and some commands.
    return[0]: the commands to run
    return[1]: the `BoolVar`
    """
    if isinstance(expr, BoolVar):
        return [], expr
    else:
        var = BoolVar.new(compiler, tmp=tmp)
        return expr.export(var, compiler), var



"""Values of "type" type in Acacia.
e.g. "int" is a "type".
"""

__all__ = ['TypeDataType', 'Type']

from typing import TYPE_CHECKING
from abc import ABCMeta, abstractmethod

from acaciamc.error import *
from acaciamc.mccmdgen.expr import *
from acaciamc.mccmdgen.datatype import DefaultDataType
from acaciamc.mccmdgen.ctexpr import CTDataType, CTCallable

if TYPE_CHECKING:
    from acaciamc.mccmdgen.datatype import DataType

class TypeDataType(DefaultDataType):
    name = "type"

ctdt_type = CTDataType("type")

class Type(ConstExprCombined, AcaciaCallable, CTCallable, metaclass=ABCMeta):
    """Base class for type of a variable that represents a type.
    (e.g. type of builtin "int" is `IntType`).
    """
    cdata_type = ctdt_type

    def __init__(self):
        super().__init__(TypeDataType())
        self.func_repr = str(self.datatype_hook())
        self.do_init()

    def do_init(self):
        """Initialzer for `Type`s. This exists due to historical reason."""
        pass

    def call(self, args, keywords, compiler):
        """
        Calling a type is to create an instance of this type,
        i.e. calling `__new__`.
        """
        new = self.attribute_table.lookup('__new__')
        if new is None or not isinstance(new, AcaciaCallable):
            raise Error(ErrorType.CANT_CREATE_INSTANCE,
                        type_=str(self.datatype_hook()))
        return new.call(args, keywords, compiler)

    def ccall(self, args, keywords, compiler):
        new = self.attributes.clookup('__new__')
        if new is None or not isinstance(new, CTCallable):
            raise Error(ErrorType.CANT_CREATE_INSTANCE,
                        type_=self.cdatatype_hook().name)
        return new.ccall(args, keywords, compiler)

    @abstractmethod
    def datatype_hook(self) -> "DataType":
        pass

    @abstractmethod
    def cdatatype_hook(self) -> "CTDataType":
        pass



"""Generic in Acacia."""

__all__ = ["GenericDataType", "BinaryGeneric"]

from acaciamc.mccmdgen.datatype import DefaultDataType
from acaciamc.mccmdgen.ctexpr import CTDataType
from acaciamc.mccmdgen.expr import *

class GenericDataType(DefaultDataType):
    name = 'generic'

ctdt_generic = CTDataType("generic")

class BinaryGeneric(ConstExprCombined):
    cdata_type = ctdt_generic

    def __init__(self):
        super().__init__(GenericDataType())



"""Collections of several variables at runtime."""

__all__ = ["StructDataType", "Struct"]

from typing import TYPE_CHECKING, Dict, Tuple
from itertools import chain

from acaciamc.mccmdgen.expr import *
from acaciamc.mccmdgen.datatype import SupportsEntityField, Storable
from acaciamc.error import *

if TYPE_CHECKING:
    from .entity import _EntityBase
    from .struct_template import StructTemplate
    from .types import DataType

class StructDataType(Storable, SupportsEntityField):
    def __init__(self, template: "StructTemplate"):
        super().__init__()
        self.template = template

    def __str__(self) -> str:
        return self.template.name

    @classmethod
    def name_no_generic(self) -> str:
        return "struct"

    def matches(self, other: "DataType") -> bool:
        return (isinstance(other, StructDataType) and
                other.template.is_subtemplate_of(self.template))

    def new_var(self, compiler) -> "Struct":
        return Struct.from_template(self.template, compiler)

    def new_entity_field(self, compiler):
        field_info: Dict[str, Tuple[dict, SupportsEntityField]] = {}
        for name, type_ in self.template.field_types.items():
            if not isinstance(type_, SupportsEntityField):
                raise Error(ErrorType.UNSUPPORTED_EFIELD_IN_STRUCT,
                            template=self.template.name, field_type=type_)
            submeta = type_.new_entity_field(compiler)
            field_info[name] = (submeta, type_)
        return {"field_info": field_info}

    def new_var_as_field(
            self, entity: "_EntityBase",
            field_info: Dict[str, Tuple[dict, SupportsEntityField]]
        ) -> "Struct":
        vars_ = {}
        for name, (submeta, type_) in field_info.items():
            subvar = type_.new_var_as_field(entity, **submeta)
            vars_[name] = subvar
        return Struct(self.template, vars_)

class Struct(VarValue):
    def __init__(self, template: "StructTemplate", vars_: Dict[str, VarValue]):
        super().__init__(StructDataType(template))
        self.template = template
        self.vars = vars_
        self.attribute_table.update(vars_)

    @classmethod
    def from_template(cls, template: "StructTemplate", compiler):
        vars_ = {}
        for name, type_ in template.field_types.items():
            var = type_.new_var(compiler)
            vars_[name] = var
        return cls(template, vars_)

    def export(self, other_struct: "Struct", compiler) -> CMDLIST_T:
        return list(chain.from_iterable(
            self.vars[name].export(other_struct.vars[name], compiler)
            for name in other_struct.vars
            # `self` may have fields that `other_struct` doesn't have.
        ))

    def swap(self, other: "Struct", compiler) -> CMDLIST_T:
        return list(chain.from_iterable(
            swap_exprs(self.vars[name], other.vars[name], compiler)
            for name in self.vars
        ))



"""
schedule - Working through game ticks

The `Task` object stands for a task manager.
e.g.
    import schedule
    import print
    def foo(arg: int):
        print.tell(print.format("Foo called with arg %0", arg))
    print.tell("Start")
    task -> schedule.Task(foo, arg=10)
    task.after(20)
result:
    Start
    [1 second (20 ticks) later]
    Foo called with arg 10
Note that, when 1 request is already sent using `after` and the function
is not called yet (within 20 ticks in this example), and another `after`
request is sent, the first one is overrided.
`task.has_schedule` returns whether a request already exists.
`task._timer` (int) shows how many ticks left before the target called
and is -1 when no request exists.

`Task`s can be registered to be called when a certain area of world is
loaded (`task.on_area_loaded`, `task.on_circle_loaded`,
`task.on_tickingarea`). Note these schedules cannot be canceled and
are not tracked by `task.has_schedule`.

`register_loop` register a function to be called repeatedly.
e.g.
    import schedule
    import print
    def foo():
        print.tell("Foo Called")
    schedule.register_loop(foo, interval=2)
result:
    Foo Called
    [2 ticks later]
    Foo Called
    [2 ticks later]
    Foo Called
    ...
"""

from typing import TYPE_CHECKING

from acaciamc.objects import *
from acaciamc.mccmdgen.expr import *
from acaciamc.mccmdgen.datatype import DefaultDataType
from acaciamc.mccmdgen.ctexpr import CTDataType
from acaciamc.ast import Operator
from acaciamc.tools import axe, resultlib, method_of
import acaciamc.mccmdgen.cmds as cmds

if TYPE_CHECKING:
    from acaciamc.compiler import Compiler

class TaskDataType(DefaultDataType):
    name = "Task"

ctdt_task = CTDataType("Task")

class TaskType(Type):
    """
    Task(target: function, *args, **kwds)
    A task manager that calls the given function after a period of time.
    The `args` and `kwds` are passed to the function.
    """
    def do_init(self):
        @method_of(self, "__new__")
        @axe.chop
        @axe.arg("target", axe.Callable())
        @axe.star_arg("args", axe.AnyValue())
        @axe.kwds("kwds", axe.AnyValue())
        def _new(compiler, target, args, kwds):
            res = Task(target, args, kwds, compiler)
            return res, res.timer_reset()

    def datatype_hook(self):
        return TaskDataType()

    def cdatatype_hook(self):
        return ctdt_task

class Task(ConstExprCombined):
    cdata_type = ctdt_task

    def __init__(self, target: AcaciaCallable, other_arg, other_kw,
                 compiler: "Compiler"):
        """
        target: The function to call
        other_arg & other_kw: Arguments to pass to `target`
        """
        super().__init__(TaskDataType())
        # Define an `int` which show how many ticks left before the function
        # runs; it is -1 when no request exists.
        self.timer = IntVar.new(compiler)
        # Allocate a file to call the given function
        self.target_file = cmds.MCFunctionFile()
        compiler.add_file(self.target_file)
        _res, call_cmds = target.call_withframe(
            other_arg, other_kw, compiler, location="<schedule.Task>"
        )
        self.target_file.extend(call_cmds)

        @method_of(self, "after")
        @axe.chop
        @axe.arg("delay", IntDataType)
        def _after(compiler, delay: AcaciaExpr):
            """.after(delay: int): Run the target after `delay`"""
            commands = delay.export(self.timer, compiler)
            return resultlib.commands(commands)
        @method_of(self, "cancel")
        @axe.chop
        def _cancel(compiler):
            """.cancel(): Cancel the schedule created by `after`."""
            return resultlib.commands(self.timer_reset())
        @method_of(self, "has_schedule")
        @axe.chop
        def _has_schedule(compiler):
            """
            .has_schedule() -> bool
            Whether a schedule created by `after` is in progress.
            """
            # Just return whether timer >= 0
            return self.timer.compare(
                Operator.greater_equal, IntLiteral(0), compiler
            )
        @method_of(self, "on_area_loaded")
        @axe.chop
        @axe.arg("origin", PosDataType)
        @axe.arg("offset", PosOffsetDataType)
        def _on_area_loaded(compiler, origin: Position, offset: PosOffset):
            """
            .on_area_loaded(origin: Pos, offset: Offset)
            Run function when the given area is loaded.
            """
            return resultlib.commands([cmds.Execute(
                origin.context,
                runs=cmds.ScheduleFunction(
                    self.target_file, "on_area_loaded add ~ ~ ~ %s" % offset
                ),
            )])
        @method_of(self, "on_circle_loaded")
        @axe.chop
        @axe.arg("origin", PosDataType)
        @axe.arg("radius", axe.RangedLiteralInt(0, None))
        def _on_circle_loaded(compiler, origin: Position, radius: int):
            """
            .on_circle_loaded(origin: Pos, radius: int-literal)
            Run function when the given circle with `origin` as origin
            and `radius` as radius (chunks) is loaded.
            """
            return resultlib.commands([cmds.Execute(
                origin.context,
                runs=cmds.ScheduleFunction(
                    self.target_file,
                    "on_area_loaded add circle ~ ~ ~ %d" % radius
                )
            )])
        @method_of(self, "on_tickingarea")
        @axe.chop
        @axe.arg("name", axe.LiteralString())
        def _on_tickingarea(compiler, name: str):
            """
            .on_tickingarea(name: str)
            Run function when the given ticking area is added.
            """
            return resultlib.commands([
                cmds.ScheduleFunction(
                    self.target_file,
                    "on_area_loaded add tickingarea %s" % name
                ),
            ])
        self.attribute_table.set("_timer", self.timer)
        # Write tick.mcfunction
        compiler.file_tick.write_debug("# schedule.Task")
        compiler.file_tick.write(
            cmds.Execute(
                [cmds.ExecuteScoreMatch(self.timer.slot, "0")],
                runs=cmds.InvokeFunction(self.target_file)
            ),
            # Decrease the timer by 1
            cmds.Execute(
                [cmds.ExecuteScoreMatch(self.timer.slot, "0..")],
                runs=cmds.ScbRemoveConst(self.timer.slot, 1)
            )
        )

    def timer_reset(self) -> CMDLIST_T:
        return [cmds.ScbSetConst(self.timer.slot, -1)]

@axe.chop
@axe.arg("target", axe.Callable())
@axe.arg("interval", IntDataType, default=IntLiteral(1))
@axe.star_arg("args", axe.AnyValue())
@axe.kwds("kwds", axe.AnyValue())
def register_loop(compiler: "Compiler", target: AcaciaCallable,
                  interval: AcaciaExpr, args, kwds):
    """
    schedule.register_loop(
        target: function, interval: int-literal = 1, *args, **kwds
    )
    Call a function repeatly every `interval` ticks with `args` and `kwds`.
    """
    _res, tick_commands = target.call_withframe(
        args, kwds, compiler, location="<schedule.register_loop>"
    )
    compiler.file_tick.write_debug("# schedule.register_loop")
    # Optimization: if the interval is 1, no need for timer
    if isinstance(interval, IntLiteral) and interval.value == 1:
        compiler.file_tick.extend(tick_commands)
        return None
    # Allocate an int for timer
    timer = IntVar.new(compiler)
    # Initialize
    init_cmds = [cmds.ScbSetConst(timer.slot, 0)]
    # Tick loop
    ## Call on times up AND reset timer
    tick_commands.extend(interval.export(timer, compiler))
    compiler.file_tick.extend(
        cmds.execute(
            [cmds.ExecuteScoreMatch(timer.slot, "..0")], runs=cmd
        )
        for cmd in tick_commands
    )
    ## Decrease the timer by 1
    compiler.file_tick.write(cmds.ScbRemoveConst(timer.slot, 1))
    # Result
    return resultlib.commands(init_cmds)

def acacia_build(compiler):
    return {
        "Task": TaskType(),
        "register_loop": BinaryFunction(register_loop)
    }



# Add `acaciamc` directory to path
import os
import sys
sys.path.append(os.path.realpath(
    os.path.join(__file__, os.pardir, os.pardir)
))

from acaciamc.mccmdgen import optimizer, cmds

class MyOpt(optimizer.Optimizer):
    def entry_files(self):
        return [f1]

    max_inline_file_size = 30

    def dump(self):
        return ('\n\n'.join(
            str(file) + '\n' + file.to_str(debugging=True)
            for file in self.files
        ))

project = MyOpt("scb")
f1 = cmds.MCFunctionFile("test/path1")
f2 = cmds.MCFunctionFile("test/path2")
f3 = cmds.MCFunctionFile("test/path3")
f4 = cmds.MCFunctionFile("test/path4")
f5 = cmds.MCFunctionFile("test/path5")
project.add_file(f1)
project.add_file(f2)
project.add_file(f3)
project.add_file(f4)
project.add_file(f5)
v1 = project.allocate()
v2 = project.allocate()
v3 = project.allocate()

f1.write("test command")
f1.write(cmds.ScbSetConst(v1, 10))
f1.write(cmds.ScbOperation(cmds.ScbOp.ASSIGN, v2, v1))
# c = cmds.Execute(
#     [cmds.ExecuteScoreComp(v2, v3, cmds.ScbCompareOp.GTE)],
#     cmds.InvokeFunction(f3)
# )
c = cmds.Execute(
    [cmds.ExecuteScoreMatch(v2, "114")],
    cmds.InvokeFunction(f3)
)
f1.write(c)
f1.write(cmds.Execute(
    [cmds.ExecuteEnv("as", "@p")],
    # [cmds.ExecuteCond("entity", "@p")],
    cmds.InvokeFunction(f4)
))
f1.write(cmds.RawtextOutput(
    "tellraw @a", cmds.Rawtext([cmds.RawtextScore(v2)])
))
f1.write(cmds.ScbRandom(v2, 10, 20))

f2.write(cmds.InvokeFunction(f1))

f3.write(cmds.ScbOperation(cmds.ScbOp.ASSIGN, v1, v2))
f3.write(cmds.ScbAddConst(v2, 10))
f3.write(cmds.Execute(
    [cmds.ExecuteScoreComp(v1, v3, cmds.ScbCompareOp.GTE)],
    cmds.Cmd("test arg")
))
f3.write(cmds.ScbSetConst(v1, 20))

f4.write(cmds.Execute(
    [cmds.ExecuteCond("block", "~~~ air")],
    cmds.InvokeFunction(f5)
))

f5.write(cmds.Cmd("say cmd5: 1"))
f5.write(cmds.Cmd("say cmd5: 2"))

print(project.dump())
project.optimize()
print("\n==========================\n")
print(project.dump())



"""
Function objects.

There are several types of functions:
- AcaciaFunction: functions that are written in Acacia
- InlineFunction: functions written in Acacia that are annotated with
  `inline`
- BinaryFunction: functions that are written in Python, which are
  usually builtin functions
- BoundMethod: methods that are bound to an entity and we are sure
  which implementation we are calling
- BoundVirtualMethod: virtual/override methods that are bound to an
  entity and we are not sure which implementation we are calling.
  Example:
    entity A:
      def foo():
        pass
      virtual def bar():
        pass
    entity B extends A:
      override def bar():
        pass
    def f(a: A):
      a.foo()  # BoundMethod: must be foo in A
      upcast(a, A).bar()  # BoundMethod: must be bar in A
      a.bar()  # BoundVirtualMethod: bar in A or B?
- ConstructorFunction: a function that produces a new object. This
  exists only for optimization purposes.
    # `Entity` is a constructor, so in this code:
    x := Entity()
    # there is no temporary entity used during the construction, but
    # one entity is created and directly bound to `x`.
    # Additionally, this code:
    x = Entity()
    # is optimized too, because `Entity` can reconstruct `x` and set it
    # to a new entity.
- BinaryCTFunction: functions that are written in Python and available
  under const context and runtime context.
- BinaryCTOnlyFunction: functions that are written in Python and
  only available under const context.
- AcaciaCTFunction: functions written in Acacia that are annotated with
  `const`.
"""

__all__ = [
    # Type
    'FunctionDataType',
    # Expressions
    'AcaciaFunction', 'InlineFunction', 'BinaryFunction',
    'BoundMethod', 'BoundVirtualMethod', 'ConstructorFunction',
    'AcaciaCTFunction', 'BinaryCTFunction', 'BinaryCTOnlyFunction'
]

from typing import (
    TYPE_CHECKING, List, Dict, Union, Callable, Tuple, Optional, Generic,
    TypeVar, Any
)
from abc import abstractmethod

from acaciamc.error import *
from acaciamc.mccmdgen.datatype import DefaultDataType, Storable
from acaciamc.ast import FuncPortType
from acaciamc.mccmdgen.mcselector import MCSelector
from acaciamc.mccmdgen.ctexpr import (
    CTObj, CTObjPtr, CTDataType, CTExpr, CTCallable
)
from acaciamc.mccmdgen.expr import *
from acaciamc.mccmdgen.utils import unreachable, InvalidOpError
from acaciamc.localization import localize
import acaciamc.mccmdgen.cmds as cmds
from .none import NoneLiteral
from .entity import EntityReference

if TYPE_CHECKING:
    from acaciamc.compiler import Compiler
    from acaciamc.ast import InlineFuncData, ConstFuncData
    from acaciamc.mccmdgen.datatype import DataType
    from acaciamc.mccmdgen.generator import Generator, Context
    from .entity import _EntityBase, TaggedEntity
    from .entity_template import EntityTemplate

T = TypeVar("T")
DT = TypeVar("DT")
IT = TypeVar("IT")

class ArgumentHandler(Generic[IT, T, DT]):
    """
    A tool to match function arguments against a given definition.
    `IT`: input argument type
    `T`: the main argument type, used in output and type checking
    `DT`: the type of the data type, used in type checking
    `T` and `DT` should only accept non-None values.
    """
    def __init__(self, args: List[str], arg_types: Dict[str, Optional[DT]],
                 arg_defaults: Dict[str, Optional[T]]):
        """`args`, `arg_types` and `arg_defaults` decide the expected
        pattern.
        """
        self.args = args
        self.arg_types = arg_types
        self.arg_defaults = arg_defaults
        # Throw away arguments that have no default value in
        # `arg_defaults`.
        for arg, value in self.arg_defaults.copy().items():
            if value is None:
                del self.arg_defaults[arg]
        self.ARG_LEN = len(self.args)

    def preconvert(self, arg: str, value: IT) -> T:
        """
        Convert input arguments of `IT` type to `T` type.
        Can be omitted only if `IT` is same as `T`.
        """
        return value

    def type_check(self, arg: str, value: T, type_: DT) -> Optional[T]:
        """
        Must be implemented by subclasses to check if the `value` is of
        `type_` type. If this returns successfully with `None`, then the
        value is accepted. If this returns successfully with a value of
        `T` type, then that value is used instead as the argument value.
        If the type check fails, this should not return.
        """
        raise NotImplementedError

    def __type_check(self, arg: str, value: T, type_: Optional[DT]) -> T:
        if type_ is None:
            return value
        else:
            r = self.type_check(arg, value, type_)
            return value if r is None else r

    def match(self, args: List[IT], keywords: Dict[str, IT]) -> Dict[str, T]:
        """Match the expected pattern with given call arguments.
        Return a `dict` mapping names to argument value.
        """
        if len(args) > self.ARG_LEN:
            raise Error(ErrorType.TOO_MANY_ARGS)
        res = dict.fromkeys(self.args)
        # positioned
        for i, value in enumerate(args):
            arg = self.args[i]
            value = self.preconvert(arg, value)
            res[arg] = self.__type_check(arg, value, self.arg_types[arg])
        # keyword
        for arg, value in keywords.items():
            if arg not in self.args:
                raise Error(ErrorType.UNEXPECTED_KEYWORD_ARG, arg=arg)
            if res[arg] is not None:
                raise Error(ErrorType.ARG_MULTIPLE_VALUES, arg=arg)
            value = self.preconvert(arg, value)
            res[arg] = self.__type_check(arg, value, self.arg_types[arg])
        # if any args are missing use default if exists, else error
        for arg, value in res.copy().items():
            if value is None:
                if arg in self.arg_defaults:
                    res[arg] = self.arg_defaults[arg]
                else:
                    raise Error(ErrorType.MISSING_ARG, arg=arg)
        return res

class AcaciaArgHandler(ArgumentHandler[AcaciaExpr, AcaciaExpr, "DataType"]):
    def __init__(
        self, args: List[str],
        arg_types: Dict[str, Optional["DataType"]],
        arg_defaults: Dict[str, Optional[AcaciaExpr]],
        compiler: "Compiler"
    ):
        super().__init__(args, arg_types, arg_defaults)
        self.compiler = compiler

    def type_check(self, arg: str, value: AcaciaExpr, tp: "DataType"):
        if not tp.matches(value.data_type):
            # Try implicit type conversion
            try:
                v = value.implicitcast(tp, self.compiler)
            except InvalidOpError:
                raise Error(
                    ErrorType.WRONG_ARG_TYPE, arg=arg,
                    expect=str(tp), got=str(value.data_type)
                )
            else:
                return v

class FunctionDataType(DefaultDataType):
    name = 'function'

ctdt_function = CTDataType("function")

class AcaciaFunction(ConstExprCombined, AcaciaCallable):
    cdata_type = ctdt_function

    def __init__(self, name: str, args: List[str],
                 arg_types: Dict[str, "Storable"],
                 arg_defaults: Dict[str, Union[AcaciaExpr, None]],
                 returns: "Storable", compiler: "Compiler",
                 source=None):
        super().__init__(FunctionDataType())
        self.name = name
        self.result_type = returns
        self.arg_handler = AcaciaArgHandler(
            args, arg_types, arg_defaults, compiler
        )
        # Create a `VarValue` for every args according to their types
        # and store them as dict at `self.arg_vars`.
        self.arg_vars: Dict[str, VarValue] = {
            arg: arg_type.new_var(compiler)
            for arg, arg_type in arg_types.items()
        }
        # Allocate a var for result value
        self.result_var = returns.new_var(compiler)
        # `file`: the target file of function. When it is None,
        # the function is empty. It should be assigned by `Generator`.
        self.file: Union[cmds.MCFunctionFile, None] = None
        # For error hint
        if source is not None:
            self.source = source
        self.func_repr = self.name

    def call(self, args: ARGS_T, keywords: KEYWORDS_T, compiler) -> CALLRET_T:
        res = []
        # Parse args
        arguments = self.arg_handler.match(args, keywords)
        # Assign argument values to `arg_vars`
        for arg, value in arguments.items():
            res.extend(value.export(self.arg_vars[arg], compiler))
        # Call function
        if self.file is not None:
            res.append(cmds.InvokeFunction(self.file))
        # Make a copy of `result_var`
        result = self.result_type.new_var(compiler)
        res.extend(self.result_var.export(result, compiler))
        result.is_temporary = True
        return result, res

class InlineFunction(ConstExprCombined, AcaciaCallable):
    cdata_type = ctdt_function

    def __init__(self, name: str, node: "InlineFuncData",
                 args, arg_types, arg_defaults, arg_ports,
                 returns: "DataType", result_port: FuncPortType,
                 context: "Context", owner: "Generator",
                 source=None):
        super().__init__(FunctionDataType())
        # We store the InlineFuncData node directly
        self.node = node
        self.owner = owner
        self.context = context
        self.name = name
        self.result_type = returns
        self.result_port = result_port
        self.arg_handler = AcaciaArgHandler(
            args, arg_types, arg_defaults, owner.compiler
        )
        self.arg_ports = arg_ports
        # For error hint
        if source is not None:
            self.source = source
        self.func_repr = self.name

    def call(self, args: ARGS_T, keywords: KEYWORDS_T, compiler) -> CALLRET_T:
        return self.owner.call_inline_func(self, args, keywords)

def _handle_impl_res(res: Union[CALLRET_T, AcaciaExpr, None]):
    if isinstance(res, tuple):  # CALLRET_T
        return res
    elif isinstance(res, AcaciaExpr):
        return res, []
    elif res is None:
        return NoneLiteral(), []
    else:
        unreachable("Invalid return of binary func "
                    f"implementation: {res!r}")

class BinaryFunction(ConstExprCombined, AcaciaCallable):
    """These are the functions that are written in Python,
    rather than `AcaciaFunction`s which are written in Acacia.
    The arguments passed to `AcaicaFunction` are *assigned* to local
    vars (AcaciaFunction.arg_vars) using commands, while arguments
    passed to `BinaryFunction` are directly handled in Python and no
    command will be generated for passing the arguments (similar to
    inline functions).
    As a result, the arguments can be "unstorable" -- argument of any
    type will be accepted. Also, the result value can be "unstorable"
    -- any type of result can be returned.
    """
    cdata_type = ctdt_function

    def __init__(
        self,
        implementation: Callable[
            ["Compiler", ARGS_T, KEYWORDS_T],
            Union[CALLRET_T, AcaciaExpr, None]
        ]
    ):
        """implementation: it handles a call to this binary function.
        It should accept 3 arguments: compiler, args and keywords.
          compiler: the `Compiler` object
          args: list of positional arguments passed to this function
          keywords: dict that holds keyword arguments (keys are strings
          representing keyword name and values are argument values)
        NOTE Dealing with arguments can be annoying, BUT we provide an
        argument parsing tool called Axe (see acaciamc/tools/axe.py).
        It should return any `AcaciaExpr` as the result value, a tuple
        (element 1 is result value, element 2 is list of strings
        representing commands) or None (returns acacia None and writes
        no command).
        """
        super().__init__(FunctionDataType())
        self.implementation = implementation
        self.func_repr = localize("objects.function.binary")

    def call(self, args: ARGS_T, keywords: KEYWORDS_T, compiler) -> CALLRET_T:
        return _handle_impl_res(self.implementation(compiler, args, keywords))

class CTArgHandler(ArgumentHandler[
    Union[CTExpr, AcaciaExpr], CTExpr, "CTDataType"
]):
    def preconvert(self, arg: str, value: Union[CTExpr, AcaciaExpr]):
        if isinstance(value, AcaciaExpr):
            if not isinstance(value, ConstExpr):
                raise Error(ErrorType.ARG_NOT_CONST, arg=arg)
            value = value.to_ctexpr()
        return value

    def type_check(self, arg: str, value: CTExpr, dt: "CTDataType"):
        v = abs(value)
        if not dt.is_typeof(v):
            # Try implicit type conversion
            try:
                new_value = value.cimplicitcast(dt)
            except InvalidOpError:
                raise Error(ErrorType.WRONG_ARG_TYPE, arg=arg,
                            expect=dt.name, got=v.cdata_type.name)
            else:
                return new_value

class AcaciaCTFunction(ConstExprCombined, AcaciaCallable, CTCallable):
    cdata_type = ctdt_function

    def __init__(self, name: str, node: "ConstFuncData",
                 args, arg_types, arg_defaults,
                 returns: "CTDataType",
                 context: "Context", owner: "Generator",
                 source=None):
        super().__init__(FunctionDataType())
        self.node = node
        self.owner = owner
        self.context = context
        self.name = name
        self.result_type = returns
        self.arg_handler = CTArgHandler(args, arg_types, arg_defaults)
        # For error hint
        if source is not None:
            self.source = source
        self.func_repr = self.name

    def call(self, args: ARGS_T, keywords: KEYWORDS_T, compiler) -> CALLRET_T:
        arg2value = self.arg_handler.match(args, keywords)
        r = abs(self.owner.ccall_const_func(self, arg2value))
        try:
            return r.to_rt(), []
        except InvalidOpError:
            raise Error(ErrorType.NON_RT_RESULT, got=r.cdata_type.name)

    def ccall(self, args: List["CTObj"],
              keywords: Dict[str, "CTObj"], compiler):
        arg2value = self.arg_handler.match(args, keywords)
        return self.owner.ccall_const_func(self, arg2value)

class BinaryCTOnlyFunction(CTCallable):
    cdata_type = ctdt_function

    def __init__(
        self, impl: Callable[
            ["Compiler", List[CTExpr], Dict[str, CTExpr]],
            Union[None, CTExpr, ConstExpr]
        ],
        *args, **kwds
    ):
        super().__init__(*args, **kwds)
        self.impl = impl
        self.func_repr = localize("objects.function.binary")

    def ccall(self, args: List[CTExpr], keywords: Dict[str, CTExpr],
              compiler: "Compiler") -> CTExpr:
        res = self.impl(compiler, args, keywords)
        if res is None:
            return NoneLiteral()
        if isinstance(res, (CTObj, CTObjPtr)):
            return res
        if isinstance(res, ConstExpr):
            return res.to_ctexpr()
        unreachable("invalid return value from impl of"
                    f" BinaryCTOnlyFunction: {res!r}")

class BinaryCTFunction(BinaryCTOnlyFunction, ConstExprCombined,
                       AcaciaCallable):
    cdata_type = ctdt_function

    def __init__(
        self, impl: Callable[
            ["Compiler", List[Union[CTExpr, AcaciaExpr]],
             Dict[str, Union[CTExpr, AcaciaExpr]]],
            Union[None, CTExpr, AcaciaExpr, CALLRET_T]
        ]
    ):
        super().__init__(impl, FunctionDataType())
        self.func_repr = localize("objects.function.binary")

    def call(self, args: ARGS_T, keywords: KEYWORDS_T, compiler) -> CALLRET_T:
        res = self.impl(compiler, args, keywords)
        if isinstance(res, (CTObj, CTObjPtr)):
            res = abs(res)
            try:
                return res.to_rt(), []
            except InvalidOpError:
                raise Error(ErrorType.NON_RT_RESULT, got=res.cdata_type.name)
        return _handle_impl_res(res)

METHODDEF_T = Union[AcaciaFunction, InlineFunction]

class _BoundMethod(ConstExprCombined, AcaciaCallable):
    """
    A method bound to an entity, but when called it does not set self
    var. Self var is set by `BoundMethod` and `BoundVirtualMethod`.
    """
    cdata_type = ctdt_function

    def __init__(self, object_: "_EntityBase", method_name: str,
                 definition: METHODDEF_T):
        super().__init__(FunctionDataType())
        self.name = method_name
        self.object = object_
        self.definition = definition
        self.is_inline = isinstance(definition, InlineFunction)
        self.func_repr = self.name
        self.source = self.definition.source

    def call(self, args: ARGS_T, keywords: KEYWORDS_T, compiler) -> CALLRET_T:
        if self.is_inline:
            assert isinstance(self.definition, InlineFunction), \
                   "Unexpected target function %r" % self.definition
            # `EntityReference` is needed so that `self` is not
            # assignable.
            self.definition.context.self_value = \
                EntityReference.from_other(self.object)
        result, commands = self.definition.call(args, keywords, compiler)
        # `BinaryFunction`s cannot be method implementation
        # because we are not sure about their result data type
        if self.is_inline:
            # For gc...
            self.definition.context.self_value = None
        return result, commands

class BoundMethod(ConstExprCombined, AcaciaCallable):
    cdata_type = ctdt_function

    def __init__(self, object_: "_EntityBase", method_name: str,
                 definition: METHODDEF_T,
                 get_self_var: Optional[Callable[[], "TaggedEntity"]],
                 # Getter of self var is needed to prevent circular
                 # dependency: creating an entity requires binding
                 # methods, binding a method requires self var which is
                 # just an entity, and creating an entity ...
                 ):
        super().__init__(FunctionDataType())
        self.content = _BoundMethod(object_, method_name, definition)
        self.func_repr = self.content.func_repr
        self.source = self.content.source
        self.self_var_getter = get_self_var
        assert (get_self_var is None) == self.content.is_inline

    def call(self, args: ARGS_T, keywords: KEYWORDS_T, compiler) -> CALLRET_T:
        result, commands = self.content.call(args, keywords, compiler)
        if not self.content.is_inline:
            commands[:0] = self.content.object.export(
                self.self_var_getter(), compiler
            )
        return result, commands

class BoundVirtualMethod(ConstExprCombined, AcaciaCallable):
    cdata_type = ctdt_function

    def __init__(self, object_: "_EntityBase", method_name: str,
                 result_var: VarValue, compiler: "Compiler"):
        super().__init__(FunctionDataType())
        self.name = method_name
        self.object = object_
        self.impls: \
            Dict[METHODDEF_T, Tuple[
                _BoundMethod, List["EntityTemplate"],
                Optional[Callable[[], "TaggedEntity"]]
            ]] = {}
        self.files: \
            List[Tuple[ARGS_T, KEYWORDS_T, cmds.MCFunctionFile]] = []
        self.result_var = result_var
        self.func_repr = self.name
        compiler.before_finish(self._generate)

    def add_implementation(
        self, template: "EntityTemplate", definition: METHODDEF_T,
        get_self_var: Optional[Callable[[], "TaggedEntity"]]
    ):
        if definition in self.impls:
            _, templates, _ = self.impls[definition]
            templates.append(template)
        elif template.is_subtemplate_of(self.object.template):
            bound_method = _BoundMethod(self.object, self.name, definition)
            self.impls[definition] = (bound_method, [template], get_self_var)

    def call(self, args: ARGS_T, keywords: KEYWORDS_T, compiler) -> CALLRET_T:
        file = cmds.MCFunctionFile()
        self.files.append((args, keywords, file))
        compiler.add_file(file)
        return self.result_var, [
            cmds.Execute(
                [cmds.ExecuteEnv("as", self.object.to_str())],
                cmds.InvokeFunction(file)
            )
        ]

    def _generate(self, compiler: "Compiler"):
        def _call_bm(args: ARGS_T, keywords: KEYWORDS_T,
                     bm: _BoundMethod) -> CMDLIST_T:
            result, commands = bm.call_withframe(
                args, keywords, compiler,
                location=
                    localize("objects.function.bvm.dispatcher") % self.name
            )
            commands.extend(result.export(self.result_var, compiler))
            return commands

        if len(self.impls) == 1:
            only_bm, _, _getself = next(iter(self.impls.values()))
            only_selfvar = None if _getself is None else _getself()
        else:
            only_bm = only_selfvar = None
        for args, keywords, file in self.files:
            file.write_debug(
                "## Virtual method dispatcher for %s.%s()"
                % (self.object.template.name, self.name)
            )
            # Optimize: only one implementation
            if only_bm is not None:
                file.write_debug("# Only one implementation found")
                if only_selfvar is not None:
                    file.extend(only_selfvar.clear())
                    # XXX direct access to TaggedEntity.tag
                    file.write("tag @s add %s" % only_selfvar.tag)
                file.extend(_call_bm(args, keywords, only_bm))
                continue
            # Fallback
            for impl, (_, templates, get_self_var) in self.impls.items():
                file.write_debug(
                    "# For %s"
                    % (", ".join(template.name for template in templates))
                )
                sel = MCSelector("s")
                for template in templates:
                    sel.scores(compiler.etemplate_id_scb,
                               f"!{template.runtime_id}")
                sel_s = sel.to_str()
                commands = _call_bm(args, keywords, impl)
                if not commands:
                    continue
                f = cmds.MCFunctionFile()
                compiler.add_file(f)
                f.write_debug("## Helper for virtual method dispatcher")
                f.extend(commands)
                if get_self_var is None:
                    file.write(cmds.Execute(
                        [
                            cmds.ExecuteCond("entity", sel_s, invert=True),
                            # Make sure @s is alive:
                            cmds.ExecuteCond("entity", "@s")
                        ],
                        runs=cmds.InvokeFunction(f)
                    ))
                else:
                    self_var = get_self_var()
                    # XXX direct access to TaggedEntity.tag
                    self_tag = self_var.tag
                    file.extend(self_var.clear())
                    file.write(cmds.Execute(
                        [cmds.ExecuteCond("entity", sel_s, invert=True)],
                        runs="tag @s add %s" % self_tag
                    ))
                    file.write(cmds.Execute(
                        [cmds.ExecuteCond("entity", "@s[tag=%s]" % self_tag)],
                        runs=cmds.InvokeFunction(f)
                    ))

class ConstructorFunction(AcaciaCallable):
    def call(self, args: ARGS_T, keywords: KEYWORDS_T, compiler) -> CALLRET_T:
        dt, kwds = self.pre_initialize(args, keywords, compiler)
        instance = dt.new_var(compiler)
        commands = self.initialize(instance, compiler, **kwds)
        return instance, commands

    @abstractmethod
    def initialize(self, instance, compiler, **kwds) -> CMDLIST_T:
        """Initialize the `instance` and return commands to execute."""
        pass

    def pre_initialize(self, args: ARGS_T, keywords: KEYWORDS_T, compiler) \
            -> Tuple[Storable, Dict[str, Any]]:
        """
        Called before `initialize` to determine the type of instance and
        which arguments to pass to `initialize`.
        """
        try:
            dt = self.datatype_hook()
            if not isinstance(dt, Storable):
                raise InvalidOpError
        except InvalidOpError:
            raise NotImplementedError(
                'pre_initialize or datatype_hook must be implemented'
            )
        return dt, {"args": args, "keywords": keywords}



"""
`ExternEngroup` objects.

When dealing with entities that exist already in the world, how can
Acacia know which template to use? Well, `ExternEngroup` allows you to
tell Acacia explicitly which template to use for those entities.
    import world
    entity MyTemp:
        pass
    entity Diamond:
        pass
    ex := ExternEngroup()
    ex2 := ExternEngroup()
    ex.select(Enfilter().is_type("armor_stand").is_name("foo"))
    for e in ex:
        # Here `e` is an entity with a special template that is not even
        # a subtemplate of `Entity`.
        if world.is_block(Pos(e).offset(y=-1), "diamond_block"):
            ex2.add(e)
            ex.remove(e)
In the example we select all armor stands with name "foo" and add
them to `ex2` if they are on diamond block, or else to `ex`.
    # Mark all current entities in `ex` as `MyTemp`, and get the
    # resolved `Engroup[MyTemp]`.
    my_group := ex.resolve(MyTemp)
    diamond_group := ex2.resolve(Diamond)
Now `my_group` is of type `Engroup[MyTemp]` and `diamond_group` is
of type `Engroup[Diamond]`.
Selecting entities that already have a template using an
`ExternEngroup` is allowed, but just make sure no entity is registered
in 2 different templates. Registering the same entity on exact same
template is allowed, though.
"""

__all__ = ["ExternEGroupDataType", "ExternEGroupType", "ExternEGroup"]

from typing import TYPE_CHECKING

from acaciamc.tools import axe
from acaciamc.mccmdgen import cmds
from acaciamc.mccmdgen.expr import *
from acaciamc.localization import localize
from .entity_group import *
from .entity_template import ETemplateDataType
from .functions import (
    ConstructorFunction, FunctionDataType, BinaryFunction, ctdt_function
)

if TYPE_CHECKING:
    from acaciamc.compiler import Compiler
    from .entity_template import EntityTemplate

class ExternEGroupDataType(EGroupDataType):
    def __init__(self, compiler: "Compiler"):
        super().__init__(compiler.external_template)

    def __str__(self) -> str:
        return "ExternEngroup"

    @classmethod
    def name_no_generic(cls) -> str:
        return "ExternEngroup"

    def new_var(self, compiler):
        return ExternEGroup(compiler)

class ExternEGroupType(EGroupType):
    def __init__(self, compiler: "Compiler"):
        super().__init__(compiler.external_template)
        self.compiler = compiler

    def datatype_hook(self):
        return ExternEGroupDataType(self.compiler)

class _ExternEGroupResolve(ConstExprCombined, ConstructorFunction):
    cdata_type = ctdt_function

    def __init__(self, owner: "ExternEGroup"):
        super().__init__(FunctionDataType())
        self.owner = owner
        self.func_repr = (localize("objects.externegroup.resolve")
                          % self.owner.data_type)

    def initialize(self, instance: "EntityGroup", compiler: "Compiler"):
        SELF = self.owner.get_selector().to_str()
        template = instance.template
        commands = EGroupType(template).initialize(instance, compiler, [], {})
        commands.append(cmds.ScbSetConst(
            cmds.ScbSlot(SELF, compiler.etemplate_id_scb),
            template.runtime_id
        ))
        commands.append("tag %s add %s" % (SELF, instance.tag))
        return commands

    def pre_initialize(self, args: ARGS_T, keywords: KEYWORDS_T, compiler):
        template_out = None
        @axe.chop
        @axe.arg("template", ETemplateDataType)
        def _call_me(compiler, template: "EntityTemplate"):
            nonlocal template_out
            template_out = template
        BinaryFunction(_call_me).call(args, keywords, compiler)
        return EGroupDataType(template_out), {}

class ExternEGroup(EntityGroup):
    def __init__(self, compiler: "Compiler"):
        super().__init__(ExternEGroupDataType(compiler), compiler)
        self.attribute_table.set("resolve", _ExternEGroupResolve(self))



"""Acacia tools for creating binary modules."""

__all__ = ["method_of", "cmethod_of", "ImmutableMixin", "transform_immutable"]

from typing import Callable as _Callable, Union as _Union

# `import acaciamc.objects as objects` won't work in 3.6
# because of a Python bug (see https://bugs.python.org/issue23203)
from acaciamc import objects
from acaciamc.mccmdgen import ctexpr, expr

def method_of(obj: "expr.AcaciaExpr", name: str):
    """Return a decorator that defines a method for `obj` with `name`
    whose implementation is decorated function.
    """
    def _decorator(func: _Callable):
        obj.attribute_table.set(name, objects.BinaryFunction(func))
        return func
    return _decorator

def cmethod_of(obj: _Union[expr.AcaciaExpr, ctexpr.CTObj],
               name: str, runtime=True):
    table1 = table2 = None
    if isinstance(obj, expr.AcaciaExpr):
        table1 = obj.attribute_table
    if isinstance(obj, ctexpr.CTObj):
        table2 = obj.attributes
    def _decorator(func):
        if runtime:
            obj = objects.BinaryCTFunction(func)
        else:
            obj = objects.BinaryCTOnlyFunction(func)
        if table1:
            table1.set(name, obj)
        if table2:
            table2.set(name, obj)
        return func
    return _decorator

class ImmutableMixin:
    """An `AcaciaExpr` that can't be changed and only allows
    transformations into another object of same type.
    """
    def copy(self) -> "ImmutableMixin":
        raise NotImplementedError

def transform_immutable(self: ImmutableMixin):
    """Return a decorator for binary function implementations that
    transforms an immutable expression to another one.
    """
    if not isinstance(self, ImmutableMixin):
        raise TypeError("can't transform non-ImmutableMixin")
    def _decorator(func: _Callable):
        def _decorated(*args, **kwds):
            return func(self.copy(), *args, **kwds)
        return _decorated
    return _decorator



"""
world - Interact with the Minecraft world.

Types:
Item: str (item id) or Item object (id + data + components)
Block: str (block id) or Block object (id + block states)
Selector: entity or Engroup
PlayerSelector: player type selector

TODO list:
Block related:
/clone
clone(origin: Pos, offset: Offset, dest: Offset,
      mode: "move" | "force" | "replace", no_air=False)
clone_only(origin: Pos, offset: Offset, dest: Offset,
           block: Block, mode: "move" | "force" | "replace")
/fill
fill(origin: Pos, offset: Offset, block: Block,
     replacement: "destroy" | "hollow" | "outline" | "aironly" | "all")
fill_replace(origin: Pos, offset: Offset, block: Block,
             replaces: Block)
/setblock
setblock(pos: Pos, block: Block,
         replacement: "aironly" | "destroy" | "all")

Entity related:
/damage
damage(target: Selector, amount: int-literal, cause: None | str,
       damager: None | entity)
/effect
effect_give(target: Selector, effect: str, duration: int-literal,
            amplifier: int-literal = 0, particle=True)
effect_clear(target: Selector)
/enchant
enchant(target: Selector, enchantment: str, level: int-literal = 1)
/event
event_entity(target: Selector, event: str)
/kill
kill(target: Selector)
/me
msg_me(sender: entity, message: str)
/replaceitem
replaceitem_block(pos: Pos, slot: int-literal, item: Item,
                  amount=1, keep_old=False)
replaceitem_entity(target: Selector, location: str, slot: int-literal,
                   item: Item, amount=1, keep_old=False)
/ride
ride_start_tp_ride(rider: entity, ride: entity)
ride_start_tp_rider(rider: Selector, ride: entity,
                    fill: "if_group_fits" | "until_full" = "until_full")
ride_stop(rider: Selector)
ride_evict_riders(ride: Selector)
summon_rider(ride: Selector, type: str, event: None | str = None,
             name: None | str)
summon_ride(rider: Selector, type: str, event: None | str = None,
            name: None | str, mode: "skip_riders" | "no_ride_change"
            | "reassign_rides" = "reassign_rides")
/say
msg_say(sender: entity, message: str)
/spreadplayers
spread(target: Selector, center: Pos, range: float, interval=0.0)
/summon (pending)
# Consider using Acacia's entity system first.
summon(type: str, pos: Pos,
       event: None | str = None, name: None | str = None,
       rot [1.19.70+]: Rot = Rot(0, 0))
/tag (pending)
# Consider using Acacia's entity system first.
tag_add(target: Selector, tag: str)
tag_remove(target: Selector, tag: str)
/tell
msg_tell(sender: entity, receiver: Selector, message: str)
/tp
tp(target: Selector, dest: Pos, check_for_blocks=False)
rotate(target: Selector, rot: Rot)
# Just for convenience:
move(target: Selector, x=0.0, y=0.0, z=0.0, check_for_blocks=False)
move_local(target: Selector, left=0.0, up=0.0, front=0.0,
           check_for_blocks=False)

Player only:
/ability (EDU)
ability(player: PlayerSelector, ability: str, value: bool-literal)
/clear
clear_inventory(player: PlayerSelector)
clear(player: PlayerSelector, item_id: str, item_data: int-literal = -1,
      max_count: int-literal = -1)
/clearspawnpoint
spawnpoint_clear(player: PlayerSelector)
/gamemode
gamemode(player: PlayerSelector, mode: "survival" | "creative"
         | "adventure" | "default" | "spectator")
/give
give(player: PlayerSelector, item: Item, amount=1)
/kick
kick(player_name: str, reason: str | None = None)
/recipe
recipe(player: PlayerSelector, recipe: str, unlock: bool-literal)
/spawnpoint
spawnpoint_set(player: PlayerSelector, pos: Pos)
/xp
xp_add_levels(player: PlayerSelector, amount: int-literal)
xp_add_points(player: PlayerSelector, amount: int-literal)

Client side:
/camera (WIP)
camera_clear(player: Selector)
camera_fade(player: Selector,
            red: int-literal | None,
            green: int-literal | None,
            blue: int-literal | None,
            fade_in: float | None,
            hold: float | None,
            fade_out: float | None)
camera_set(player: Selector, preset: str,
           pos: Pos | None, rot: Rot | None,
           ease_time: float | None, ease_type: str | None)
/camerashake
camerashake_add(player: Selector, intensity: float,
                seconds: float, type: "positional" | "rotational")
camerashake_stop(player: Selector)
/dialogue
dialogue_open(npc: entity, player: Selector, scene: str | None = None)
dialogue_change(npc: entity, scene: str,
                player: Selector = <All players>)
/fog
fog_remove(player: Selector, name: str)
fog_pop(player: Selector, name: str)
fog_push(player: Selector, fog: str, name: str | None = None) -> str
/particle
particle(particle: str, pos: Pos)
/playanimation
playanimation(target: Selector, animation: str, next_state="default",
              fade_out=0.0, stop_molang="query.any_animation_finished",
              controller: str | None)
/playsound
sound_play(player: Selector, sound: str, pos: Pos, volume=1.0,
           pitch=1.0, min_volume=0.0)
/stopsound
sound_stop(player: Selector, sound: str | None = None)

Global:
/difficulty
settings_difficulty(value: "easy" | "normal" | "hard" | "peaceful")
/function
# Just for calling other works that have been done in mcfunction.
function(function: str)
/gamerule
settings(name: str, value: bool-literal | int-literal)
/gametest (?, pending)
/immutableworld (EDU)
settings_immutableworld(value: bool-literal)
/mobevent
settings_mobevent_master(value: bool-literal)
settings_mobevent(event: str, value: bool-literal)
/music
music_play(track: str, volume=1.0, fade=0.0, repeat=False)
music_queue(track: str, volume=1.0, fade=0.0, repeat=False)
music_stop(fade=0.0)
music_volume(volume: float)
/scoreboard (pending)
# Maybe add a module for displaying scoreboard instead of these?
scoreboard_add_objective(display_name: None | str = None,
                         name: None | str = None) -> str
scoreboard_remove_objective(name: str)
scoreboard_display(name: str, slot: "sidebar" | "list" | "belowname",
                   order: "ascending" | "descending")
scoreboard_display_clear(slot: "sidebar" | "list" | "belowname")
/script (?, pending)
/scriptevent (?)
scriptevent(message_id: str, message: str)
/setworldspawn
spawnpoint_world(pos: Pos)
/tickingarea
tickingarea_add(origin: Pos, offset: Offset, preload=False,
                name: str | None = None) -> str
tickingarea_add_circle(center: Pos, radius: int-literal, preload=False,
                       name: str | None = None) -> str
tickingarea_remove(name: str)
tickingarea_remove_pos(pos: Pos)
tickingarea_remove_all()
tickingarea_preload(name: str, state: bool-literal)
tickingarea_preload_pos(pos: Pos, state: bool-literal)
/time
time_add(ticks: int-literal)
time_set(value: int-literal | "day" | "night" | "noon"
                | "midnight" | "sunrise" | "sunset")
/toggledownfall
weather_toggle()
/volumearea (?)
volumearea_add(id: str, origin: Pos, offset: Offset,
               name: str | None = None) -> str
volumearea_remove(name: str)
volumearea_remove_pos(pos: Pos)
volumearea_remove_all()
/weather
weather(weather: "clear" | "rain" | "thunder",
        duration: int-literal | None = None)
/wsserver
wsserver_connect(url: str)
wsserver_out()

Miscellaneous:
/loot
# Loot object
Loot.kill(target: entity, weapon: str) -> Loot
Loot.kill_hand(target: entity, killer: entity,
               hand: "mainhand" | "offhand") -> Loot
Loot(table: str, tool: str) -> Loot
Loot.hand(table: str, creator: entity,
          hand: "mainhand" | "offhand") -> Loot
# Use objects
loot.to_container(pos: Pos)
loot.give(player: Selector)
loot.spawn(pos: Pos)
loot.replace_block(pos: Pos, slot: int-literal, slot_num: int-literal)
loot.replace_entity(target: Selector, location: str,
                    slot: int-literal, slot_num: int-literal)
/structure
structure_save(origin: Pos, offset: Offset,
               location: "disk" | "memory" = "memory",
               entities=True, blocks=True,
               name: str | None = None) -> str
structure_load(name: str, dest: Pos,
               rotation: 0 | 90 | 180 | 270,
               mirror: "x" | "z" | "xz" | "none",
               entities=True, blocks=True, waterlog=False,
               integrity=1.0, seed: str | None = None,
               animation: "block_by_block" | "layer_by_layer"
               | None = None, animation_seconds: float | None = None)
structure_delete(name: str)

Judgement:
is_block(pos: Pos, block: Block) -> bool
is_same_area(origin: Pos, offset: Offset, other: Offset) -> bool
is_entity(ent: entity, filter: Enfilter) -> bool
"""

from typing import Dict, Callable, List, Union, Optional, TYPE_CHECKING
import json

from acaciamc.objects import *
from acaciamc.mccmdgen.expr import *
from acaciamc.mccmdgen.datatype import DefaultDataType
from acaciamc.mccmdgen.ctexpr import CTDataType
from acaciamc.tools import axe, resultlib, cmethod_of
from acaciamc.tools.versionlib import edu_only
from acaciamc.localization import localize
import acaciamc.mccmdgen.cmds as cmds

if TYPE_CHECKING:
    from acaciamc.compiler import Compiler
    from acaciamc.mccmdgen.mcselector import MCSelector
    from acaciamc.objects.entity import _EntityBase

_methods: Dict[str, Callable] = {}

def _register(name: str):
    def _decorator(func: Callable):
        _methods[name] = func
        return func
    return _decorator

def _fmt_bool(value: bool):
    return "true" if value else "false"

class ItemDataType(DefaultDataType):
    name = "Item"

ctdt_item = CTDataType("Item")

class ItemType(Type):
    def do_init(self):
        @cmethod_of(self, "__new__")
        @axe.chop
        @axe.arg("id", axe.LiteralString(), rename="id_")
        @axe.arg("data", axe.RangedLiteralInt(0, 32767), default=0)
        @axe.arg("keep_on_death", axe.LiteralBool(), default=False)
        @axe.arg("can_destroy", axe.ListOf(axe.LiteralString()), default=[])
        @axe.arg("can_place_on", axe.ListOf(axe.LiteralString()), default=[])
        @axe.arg("lock", axe.Nullable(
            axe.LiteralStringEnum("lock_in_slot", "lock_in_inventory")
        ), default=None)
        def _new(compiler, id_: str, data: int, keep_on_death: bool,
                 can_destroy: List[str], can_place_on: List[str], lock: str):
            components = {}
            if can_destroy:
                components["minecraft:can_destroy"] = {"blocks": can_destroy}
            if can_place_on:
                components["minecraft:can_place_on"] = {"blocks": can_place_on}
            if keep_on_death:
                components["minecraft:keep_on_death"] = {}
            if lock:
                components["minecraft:item_lock"] = {"mode": lock}
            return Item(id_, data, components)

    def datatype_hook(self):
        return ItemDataType()

    def cdatatype_hook(self):
        return ctdt_item

class Item(ConstExprCombined):
    cdata_type = ctdt_item

    def __init__(self, id_: str, data: int, components: dict):
        super().__init__(ItemDataType())
        self.id = id_
        self.data = data
        self.components = components

    def to_str(self, fmt: str) -> str:
        c = "" if not self.components else " " + json.dumps(self.components)
        return fmt.format(id=self.id, data=self.data, components=c)

class BlockDataType(DefaultDataType):
    name = "Block"

ctdt_block = CTDataType("Block")

class BlockType(Type):
    def do_init(self):
        @cmethod_of(self, "__new__")
        @axe.chop
        @axe.arg("id", axe.LiteralString(), rename="id_")
        @axe.arg("states", axe.MapOf(
            axe.LiteralString(),
            axe.AnyOf(axe.LiteralInt(), axe.LiteralBool(), axe.LiteralString())
        ), default={})
        def _new(compiler, id_: str, states: Dict[str, Union[str, int, bool]]):
            return Block(id_, states)

    def datatype_hook(self):
        return BlockDataType()

    def cdatatype_hook(self):
        return ctdt_block

class Block(ConstExprCombined):
    cdata_type = ctdt_block

    def __init__(self, id_: str,
                 states: Dict[str, Union[str, int, bool]]):
        super().__init__(BlockDataType())
        self.id = id_
        self.states = states

    @staticmethod
    def format_bs_value(value: Union[str, int, bool]) -> str:
        if isinstance(value, bool):
            return _fmt_bool(value)
        elif isinstance(value, int):
            return str(value)
        else:
            assert isinstance(value, str)
            return '"%s"' % value

    def to_str(self, compiler: "Compiler", fmt: str = "{id} {states}") -> str:
        if compiler.cfg.mc_version >= (1, 20, 10):
            EQ = "="
        else:
            EQ = ":"
        return fmt.format(id=self.id, states="[%s]" % ",".join(
            "".join(('"%s"' % key, EQ, self.format_bs_value(value)))
            for key, value in self.states.items()
        ))

class ArgBlock(axe.AnyOf):
    def __init__(self):
        super().__init__(axe.Typed(BlockDataType), axe.LiteralString())

    def convert(self, origin: AcaciaExpr) -> Block:
        res = super().convert(origin)
        if isinstance(res, str):
            res = Block(res, {})
        return res

class ArgItem(axe.AnyOf):
    def __init__(self):
        super().__init__(axe.Typed(ItemDataType), axe.LiteralString())

    def convert(self, origin: AcaciaExpr) -> Item:
        res = super().convert(origin)
        if isinstance(res, str):
            res = Item(res, 0, {})
        return res

##### Block related #####

@_register("clone")
@axe.chop
@axe.arg("origin", PosDataType)
@axe.arg("offset", PosOffsetDataType)
@axe.arg("dest", PosOffsetDataType)
@axe.arg("mode", axe.Nullable(
    axe.LiteralStringEnum("move", "force")
), default=None)
@axe.arg("no_air", axe.LiteralBool(), default=False)
def clone(compiler: "Compiler", origin: Position, offset: PosOffset,
          dest: PosOffset, mode: Optional[str], no_air: bool):
    if mode is None:
        mode = "normal"
    mask = "masked" if no_air else "replace"
    cmd = cmds.Execute(
        origin.context,
        runs="clone ~ ~ ~ %s %s %s %s" % (offset, dest, mask, mode)
    )
    return resultlib.commands([cmd])

@_register("clone_only")
@axe.chop
@axe.arg("origin", PosDataType)
@axe.arg("offset", PosOffsetDataType)
@axe.arg("dest", PosOffsetDataType)
@axe.arg("block", ArgBlock())
@axe.arg("mode", axe.Nullable(
    axe.LiteralStringEnum("move", "force")
), default=None)
def clone_only(compiler, origin: Position, offset: PosOffset, dest: PosOffset,
               mode: Optional[str], block: Block):
    if mode is None:
        mode = "normal"
    cmd = cmds.Execute(
        origin.context, runs="clone ~ ~ ~ %s %s filtered %s %s" % (
            offset, dest, mode, block.to_str(compiler)
        )
    )
    return resultlib.commands([cmd])

@_register("fill")
@axe.chop
@axe.arg("origin", PosDataType)
@axe.arg("offset", PosOffsetDataType)
@axe.arg("block", ArgBlock())
@axe.arg("replacement", axe.LiteralStringEnum(
    "destroy", "hollow", "outline", "aironly", "all"
), default="all")
def fill(compiler, origin: Position, offset: PosOffset, block: Block,
         replacement: str):
    if replacement == "aironly":
        replacement = "keep"
    elif replacement == "all":
        replacement = "replace"
    cmd = cmds.Execute(
        origin.context, runs="fill ~ ~ ~ %s %s %s" % (
            offset, block.to_str(compiler), replacement
        )
    )
    return resultlib.commands([cmd])

@_register("fill_replace")
@axe.chop
@axe.arg("origin", PosDataType)
@axe.arg("offset", PosOffsetDataType)
@axe.arg("block", ArgBlock())
@axe.arg("replaces", ArgBlock())
def fill_replace(compiler, origin: Position, offset: PosOffset, block: Block,
         replaces: Block):
    cmd = cmds.Execute(
        origin.context, runs="fill ~ ~ ~ %s %s replace %s" % (
            offset, block.to_str(compiler), replaces.to_str(compiler)
        )
    )
    return resultlib.commands([cmd])

@_register("setblock")
@axe.chop
@axe.arg("pos", PosDataType)
@axe.arg("block", ArgBlock())
@axe.arg("replacement", axe.LiteralStringEnum(
    "aironly", "destroy", "all"
), default="all")
def setblock(compiler, pos: Position, block: Block, replacement: str):
    if replacement == "aironly":
        replacement = "keep"
    elif replacement == "all":
        replacement = "replace"
    cmd = cmds.Execute(
        pos.context, runs="setblock ~ ~ ~ %s %s" % (
            block.to_str(compiler), replacement
        )
    )
    return resultlib.commands([cmd])

##### Entity related #####

@_register("damage")
@axe.chop
@axe.arg("target", axe.Selector())
@axe.arg("amount", axe.RangedLiteralInt(0, None))
@axe.arg("cause", axe.Nullable(axe.LiteralString()), default=None)
@axe.arg("damager", axe.Nullable(axe.Typed(EntityDataType)), default=None)
def damage(compiler, target: "MCSelector", amount: int,
        cause: Optional[str], damager: Optional["_EntityBase"]):
    if damager is not None:
        if cause is None:
            raise axe.ArgumentError(
                "cause", localize("modules.world.damage.damagerpresent")
            )
        suffix = " %s entity %s" % (cause, damager)
    elif cause is not None:
        suffix = " " + cause
    else:
        suffix = ""
    cmd = "damage %s %d%s" % (target.to_str(), amount, suffix)
    return resultlib.commands([cmd])

@_register("effect_give")
@axe.chop
@axe.arg("target", axe.Selector())
@axe.arg("effect", axe.LiteralString())
@axe.arg("duration", axe.RangedLiteralInt(0, None))
@axe.arg("amplifier", axe.RangedLiteralInt(0, 255), default=0)
@axe.arg("particle", axe.LiteralBool(), default=True)
def effect_give(compiler, target: "MCSelector", effect: str,
                duration: int, amplifier: int, particle: bool):
    cmd = "effect %s %s %d %d %s" % (
        target.to_str(), effect, duration, amplifier, _fmt_bool(not particle)
    )
    return resultlib.commands([cmd])

@_register("effect_clear")
@axe.chop
@axe.arg("target", axe.Selector())
def effect_clear(compiler, target: "MCSelector"):
    return resultlib.commands(["effect %s clear" % target.to_str()])

@_register("enchant")
@axe.chop
@axe.arg("target", axe.Selector())
@axe.arg("enchantment", axe.LiteralString())
@axe.arg("level", axe.RangedLiteralInt(1, None), default=1)
def enchant(compiler, target: "MCSelector", enchantment: str, level: int):
    cmd = "enchant %s %s %d" % (target.to_str(), enchantment, level)
    return resultlib.commands([cmd])

@_register("event_entity")
@axe.chop
@axe.arg("target", axe.Selector())
@axe.arg("event", axe.LiteralString())
def event_entity(compiler, target: "MCSelector", event: str):
    cmd = "event entity %s %s" % (target.to_str(), event)
    return resultlib.commands([cmd])

@_register("kill")
@axe.chop
@axe.arg("target", axe.Selector())
def kill(compiler, target: "MCSelector"):
    cmd = "kill %s" % target.to_str()
    return resultlib.commands([cmd])

@_register("msg_me")
@axe.chop
@axe.arg("sender", EntityDataType)
@axe.arg("message", axe.LiteralString())
def msg_me(compiler, sender: "_EntityBase", message: str):
    cmd = cmds.Execute(
        [cmds.ExecuteEnv("as", sender.to_str())],
        runs="me %s" % message
    )
    return resultlib.commands([cmd])

@_register("replaceitem_block")
@axe.chop
@axe.arg("pos", PosDataType)
@axe.arg("slot", axe.LiteralInt())
@axe.arg("item", ArgItem())
@axe.arg("amount", axe.RangedLiteralInt(1, None), default=1)
@axe.arg("keep_old", axe.LiteralBool(), default=False)
def replaceitem_block(compiler, pos: Position, slot: int, item: Item,
                      amount: int, keep_old: bool):
    cmd = cmds.Execute(
        pos.context,
        "replaceitem block ~ ~ ~ slot.container %d %s %s" % (
            slot, "keep" if keep_old else "destroy",
            item.to_str("{id} %s {data}{components}" % amount)
        )
    )
    return resultlib.commands([cmd])

@_register("replaceitem_entity")
@axe.chop
@axe.arg("target", axe.Selector())
@axe.arg("location", axe.LiteralString())
@axe.arg("slot", axe.LiteralInt())
@axe.arg("item", ArgItem())
@axe.arg("amount", axe.RangedLiteralInt(1, None), default=1)
@axe.arg("keep_old", axe.LiteralBool(), default=False)
def replaceitem_entity(compiler, target: "MCSelector", location: str,
                       slot: int, item: Item, amount: int, keep_old: bool):
    cmd = "replaceitem entity %s %s %d %s %s" % (
        target.to_str(), location, slot, "keep" if keep_old else "destroy",
        item.to_str("{id} %s {data}{components}" % amount)
    )
    return resultlib.commands([cmd])

@_register("ride_start_tp_ride")
@axe.chop
@axe.arg("rider", EntityDataType)
@axe.arg("ride", EntityDataType)
def ride_start_tp_ride(compiler, rider: "_EntityBase", ride: "_EntityBase"):
    cmd = "ride %s start_riding %s teleport_ride" % (rider, ride)
    return resultlib.commands([cmd])

@_register("ride_start_tp_rider")
@axe.chop
@axe.arg("rider", axe.Selector())
@axe.arg("ride", EntityDataType)
@axe.arg("fill", axe.LiteralStringEnum("if_group_fits", "until_full"),
         default="until_full")
def ride_start_tp_rider(compiler, rider: "MCSelector",
                        ride: "_EntityBase", fill: str):
    cmd = "ride %s start_riding %s teleport_rider %s" % (
        rider.to_str(), ride, fill
    )
    return resultlib.commands([cmd])

@_register("ride_stop")
@axe.chop
@axe.arg("rider", axe.Selector())
def ride_stop(compiler, rider: "MCSelector"):
    return resultlib.commands(["ride %s stop_riding" % rider.to_str()])

@_register("ride_evict_riders")
@axe.chop
@axe.arg("ride", axe.Selector())
def ride_evict_riders(compiler, ride: "MCSelector"):
    return resultlib.commands(["ride %s evict_riders" % ride.to_str()])

@_register("summon_rider")
@axe.chop
@axe.arg("ride", axe.Selector())
@axe.arg("type", axe.LiteralString(), rename="type_")
@axe.arg("event", axe.Nullable(axe.LiteralString()), default=None)
@axe.arg("name", axe.Nullable(axe.LiteralString()), default=None)
def summon_rider(compiler, ride: "MCSelector", type_: str,
                 event: Optional[str], name: Optional[str]):
    if name is None:
        suffix = ""
    else:
        suffix = " %s" % cmds.mc_str(name)
    cmd = "ride %s summon_rider %s %s%s" % (
        ride.to_str(), type_, "*" if event is None else event, suffix
    )
    return resultlib.commands([cmd])

@_register("summon_ride")
@axe.chop
@axe.arg("rider", axe.Selector())
@axe.arg("type", axe.LiteralString(), rename="type_")
@axe.arg("event", axe.Nullable(axe.LiteralString()), default=None)
@axe.arg("name", axe.Nullable(axe.LiteralString()), default=None)
@axe.arg("mode", axe.LiteralStringEnum(
    "skip_riders", "no_ride_change", "reassign_rides"
), default="reassign_rides")
def summon_ride(compiler, rider: "MCSelector", type_: str,
                event: Optional[str], name: Optional[str], mode: str):
    if name is None:
        suffix = ""
    else:
        suffix = " %s" % cmds.mc_str(name)
    cmd = "ride %s summon_ride %s %s %s %s" % (
        rider.to_str(), type_, mode, "*" if event is None else event, suffix
    )
    return resultlib.commands([cmd])

@_register("msg_say")
@axe.chop
@axe.arg("sender", EntityDataType)
@axe.arg("message", axe.LiteralString())
def msg_say(compiler, sender: "_EntityBase", message: str):
    cmd = cmds.Execute(
        [cmds.ExecuteEnv("as", sender.to_str())],
        "say %s" % message
    )
    return resultlib.commands([cmd])

@_register("spread")
@axe.chop
@axe.arg("target", axe.Selector())
@axe.arg("center", PosDataType)
@axe.arg("range", axe.LiteralFloat(), rename="range_")
@axe.arg("interval", axe.LiteralFloat(), default=0.0)
def spread(compiler, target: "MCSelector", center: Position,
           range_: float, interval: float):
    if range_ < 1.0:
        raise axe.ArgumentError(
            "range", localize("modules.world.spread.biggerthan1")
        )
    if interval < 0.0:
        raise axe.ArgumentError(
            "interval", localize("modules.world.spread.nonnegative")
        )
    if interval + 1 > range_:
        raise axe.ArgumentError(
            "interval", localize("modules.world.spread.intervaltoobig")
        )
    cmd = cmds.Execute(
        center.context, "spreadplayers ~ ~ %s %s %s" % (
            interval, range_, target.to_str()
        )
    )
    return resultlib.commands([cmd])

@_register("summon")
@axe.chop
@axe.arg("type", axe.LiteralString(), rename="type_")
@axe.arg("pos", PosDataType)
@axe.arg("event", axe.Nullable(axe.LiteralString()), default=None)
@axe.arg("name", axe.Nullable(axe.LiteralString()), default=None)
@axe.arg("rot", RotDataType, default=None)
def summon(compiler: "Compiler", type_: str, pos: Position,
           rot: Optional[Rotation], event: Optional[str], name: Optional[str]):
    if event is None:
        event = "*"
    if name is None:
        suffix = ""
    else:
        suffix = " %s" % cmds.mc_str(name)
    if compiler.cfg.mc_version >= (1, 19, 70):
        if rot is None:
            ctx = pos.context
            rot_s = "0 0"
        else:
            ctx = pos.context + rot.context
            rot_s = "~ ~"
        cmd = cmds.Execute(
            ctx,
            runs="summon %s ~ ~ ~ %s %s%s" % (type_, rot_s, event, suffix)
        )
    else:
        if rot is not None:
            raise axe.ArgumentError(
                "rot", localize("modules.world.summon.versiontoolow")
            )
        cmd = cmds.Execute(
            pos.context, runs="summon %s ~ ~ ~ %s%s" % (type_, event, suffix)
        )
    return resultlib.commands([cmd])

@_register("tag_add")
@axe.chop
@axe.arg("target", axe.Selector())
@axe.arg("tag", axe.LiteralString())
def tag_add(compiler, target: "MCSelector", tag: str):
    cmd = "tag %s add %s" % (target.to_str(), cmds.mc_str(tag))
    return resultlib.commands([cmd])

@_register("tag_remove")
@axe.chop
@axe.arg("target", axe.Selector())
@axe.arg("tag", axe.LiteralString())
def tag_add(compiler, target: "MCSelector", tag: str):
    cmd = "tag %s remove %s" % (target.to_str(), cmds.mc_str(tag))
    return resultlib.commands([cmd])

@_register("msg_tell")
@axe.chop
@axe.arg("sender", EntityDataType)
@axe.arg("receiver", axe.Selector())
@axe.arg("message", axe.LiteralString())
def msg_tell(compiler, sender: "_EntityBase", receiver: "MCSelector",
             message: str):
    cmd = cmds.Execute(
        [cmds.ExecuteEnv("as", sender.to_str())],
        "tell %s %s" % (receiver.to_str(), message)
    )
    return resultlib.commands([cmd])

@_register("tp")
@axe.chop
@axe.arg("target", axe.Selector())
@axe.arg("dest", PosDataType)
@axe.arg("check_for_blocks", axe.LiteralBool(), default=False)
def tp(compiler, target: "MCSelector", dest: Position, check_for_blocks: bool):
    cmd = cmds.Execute(
        dest.context,
        "tp %s ~ ~ ~ %s" % (
            target.to_str(), _fmt_bool(check_for_blocks)
        )
    )
    return resultlib.commands([cmd])

@_register("rotate")
@axe.chop
@axe.arg("target", axe.Selector())
@axe.arg("rot", RotDataType)
def rotate(compiler, target: "MCSelector", rot: Rotation):
    cmd = cmds.Execute(
        [
            cmds.ExecuteEnv("as", target.to_str()),
            cmds.ExecuteEnv("at", "@s"),
            *rot.context
        ],
        "tp @s ~ ~ ~ ~ ~"
    )
    return resultlib.commands([cmd])

@_register("move")
@axe.chop
@axe.arg("target", axe.Selector())
@axe.arg("x", axe.LiteralFloat(), default=0.0)
@axe.arg("y", axe.LiteralFloat(), default=0.0)
@axe.arg("z", axe.LiteralFloat(), default=0.0)
@axe.arg("check_for_blocks", axe.LiteralBool(), default=False)
def move(compiler, target: "MCSelector", x: float, y: float, z: float,
         check_for_blocks: bool):
    cmd = cmds.Execute(
        [cmds.ExecuteEnv("as", target.to_str()),
         cmds.ExecuteEnv("at", "@s")],
        "tp @s ~%.3f ~%.3f ~%.3f %s" % (
            x, y, z, _fmt_bool(check_for_blocks)
        )
    )
    return resultlib.commands([cmd])

@_register("move_local")
@axe.chop
@axe.arg("target", axe.Selector())
@axe.arg("left", axe.LiteralFloat(), default=0.0)
@axe.arg("up", axe.LiteralFloat(), default=0.0)
@axe.arg("front", axe.LiteralFloat(), default=0.0)
@axe.arg("check_for_blocks", axe.LiteralBool(), default=False)
def move_local(compiler, target: "MCSelector", left: float, up: float,
               front: float, check_for_blocks: bool):
    cmd = cmds.Execute(
        [cmds.ExecuteEnv("as", target.to_str()),
         cmds.ExecuteEnv("at", "@s")],
        "tp @s ^%.3f ^%.3f ^%.3f %s" % (
            left, up, front, _fmt_bool(check_for_blocks)
        )
    )
    return resultlib.commands([cmd])

##### Player Only #####

@_register("ability")
@edu_only
@axe.chop
@axe.arg("player", axe.PlayerSelector())
@axe.arg("ability", axe.LiteralString())
@axe.arg("value", axe.LiteralBool())
def ability(compiler, player: "MCSelector", ability: str, value: bool):
    cmd = "ability %s %s %s" % (
        player.to_str(), ability, _fmt_bool(value)
    )
    return resultlib.commands([cmd])

@_register("clear_inventory")
@axe.chop
@axe.arg("player", axe.PlayerSelector())
def clear_inventory(compiler, player: "MCSelector"):
    return resultlib.commands(["clear %s" % player.to_str()])

@_register("clear")
@axe.chop
@axe.arg("player", axe.PlayerSelector())
@axe.arg("item_id", axe.LiteralString())
@axe.arg("item_data", axe.RangedLiteralInt(-1, None), default=-1)
@axe.arg("max_count", axe.RangedLiteralInt(-1, None), default=-1)
def clear(compiler, player: "MCSelector", item_id: str, item_data: int,
          max_count: int):
    cmd = "clear %s %s %d %d" % (
        player.to_str(), item_id, item_data, max_count
    )
    return resultlib.commands([cmd])

@_register("spawnpoint_clear")
@axe.chop
@axe.arg("player", axe.PlayerSelector())
def spawnpoint_clear(compiler, player: "MCSelector"):
    cmd = "clearspawnpoint %s" % player.to_str()
    return resultlib.commands([cmd])

@_register("gamemode")
@axe.chop
@axe.arg("player", axe.PlayerSelector())
@axe.arg("gamemode", axe.LiteralStringEnum(
    "survival", "creative", "adventure", "spectator", "default"
))
def gamemode(compiler, player: "MCSelector", gamemode: str):
    cmd = "gamemode %s %s" % (gamemode, player.to_str())
    return resultlib.commands([cmd])

@_register("give")
@axe.chop
@axe.arg("player", axe.PlayerSelector())
@axe.arg("item", ArgItem())
@axe.arg("amount", axe.RangedLiteralInt(1, 32767), default=1)
def give(compiler, player: "MCSelector", item: Item, amount: int):
    cmd = "give %s %s" % (
        player.to_str(),
        item.to_str("{id} %s {data}{components}" % amount)
    )
    return resultlib.commands([cmd])

@_register("kick")
@axe.chop
@axe.arg("player_name", axe.LiteralString())
@axe.arg("reason", axe.Nullable(axe.LiteralString()), default=None)
def kick(compiler, player_name: str, reason: Optional[str]):
    reason = "" if reason is None else " " + reason
    cmd = "kick %s%s" % (cmds.mc_str(player_name), reason)
    return resultlib.commands([cmd])

@_register("recipe")
@axe.chop
@axe.arg("player", axe.PlayerSelector())
@axe.arg("recipe", axe.LiteralString())
@axe.arg("unlock", axe.LiteralBool())
def recipe(compiler, player: "MCSelector", recipe: str, unlock: bool):
    if recipe != "*":
        recipe = cmds.mc_str(recipe)
    cmd = "recipe %s %s %s" % (
        "give" if unlock else "take",
        player.to_str(), recipe
    )
    return resultlib.commands([cmd])

@_register("spawnpoint_set")
@axe.chop
@axe.arg("player", axe.PlayerSelector())
@axe.arg("pos", PosDataType)
def spawnpoint_set(compiler, player: "MCSelector", pos: Position):
    cmd = cmds.Execute(
        pos.context,
        "spawnpoint %s ~ ~ ~" % player.to_str()
    )
    return resultlib.commands([cmd])

@_register("xp_add_levels")
@axe.chop
@axe.arg("player", axe.PlayerSelector())
@axe.arg("amount", axe.LiteralInt())
def xp_add_levels(compiler, player: "MCSelector", amount: int):
    cmd = "xp %dL %s" % (amount, player.to_str())
    return resultlib.commands([cmd])

@_register("xp_add_points")
@axe.chop
@axe.arg("player", axe.PlayerSelector())
@axe.arg("amount", axe.RangedLiteralInt(0, None))
def xp_add_points(compiler, player: "MCSelector", amount: int):
    cmd = "xp %d %s" % (amount, player.to_str())
    return resultlib.commands([cmd])

##### Client Side #####
...

##### Global #####
...

##### Misc. #####
...

##### Judgement #####

@_register("is_block")
@axe.chop
@axe.arg("pos", PosDataType)
@axe.arg("block", ArgBlock())
def is_block(compiler, pos: Position, block: Block):
    subcmds = [
        *pos.context,
        cmds.ExecuteCond("block", "~ ~ ~ %s" % block.to_str(compiler))
    ]
    return WildBool(subcmds, [])

@_register("is_same_area")
@axe.chop
@axe.arg("pos", PosDataType)
@axe.arg("offset", PosOffsetDataType)
@axe.arg("other", PosOffsetDataType)
@axe.arg("ignore_air", axe.LiteralBool(), default=False)
def is_same_area(compiler, pos: Position, offset: PosOffset,
                 other: PosOffset, ignore_air: bool):
    subcmds = [
        *pos.context,
        cmds.ExecuteCond("blocks", "~ ~ ~ %s %s %s" % (
            offset, other, "masked" if ignore_air else "all"
        ))
    ]
    return WildBool(subcmds, [])

@_register("is_entity")
@axe.chop
@axe.arg("ent", EntityDataType)
@axe.arg("filter", EFilterDataType)
def is_entity(compiler: "Compiler", ent: "_EntityBase", filter: EntityFilter):
    """
    Select entities that match the filter and return whether given
    entity is in those entities.
    """
    tmp = compiler.allocate_entity_tag()
    selector = ent.get_selector()
    selector.tag(tmp)
    commands = filter.dump("tag {selected} add %s" % tmp)
    subcmds = [cmds.ExecuteCond("entity", selector.to_str())]
    return (WildBool(subcmds, commands),
            ["tag @e[tag={0}] remove {0}".format(tmp)])

def acacia_build(compiler: "Compiler"):
    attrs = {}
    for key, value in _methods.items():
        attrs[key] = BinaryFunction(value)
    attrs["Block"] = BlockType()
    attrs["Item"] = ItemType()
    return attrs



"""Entity filter -- builtin support for selectors."""

__all__ = ["EFilterType", "EFilterDataType", "EntityFilter"]

from typing import List, Union, Optional, TYPE_CHECKING
import re

from acaciamc.error import *
from acaciamc.tools import (
    axe, versionlib,
    cmethod_of, ImmutableMixin, transform_immutable
)
from acaciamc.mccmdgen.mcselector import MCSelector, SELECTORVAR_T
from acaciamc.mccmdgen.datatype import DefaultDataType
from acaciamc.mccmdgen.ctexpr import CTDataType
from acaciamc.mccmdgen.expr import *
from acaciamc.localization import localize
import acaciamc.mccmdgen.cmds as cmds
from .types import Type
from .position import PosDataType

if TYPE_CHECKING:
    from .position import Position
    from acaciamc.mccmdgen.cmds import _ExecuteSubcmd
    from acaciamc.compiler import Compiler

RE_INT = re.compile(r"^[+-]?\s*\d+$")
PLAYER = "player"  # ID of player entity

class IntRange(axe.AnyOf):
    """Accepts a string representing integer range or an integer and
    convert it to Python string.
    """
    def __init__(self):
        super().__init__(axe.LiteralInt(), axe.LiteralString())

    def get_show_name(self) -> str:
        return localize("axe.converter.entityfilter.intrange")

    @staticmethod
    def check_int(value: str) -> bool:
        return RE_INT.match(value) is not None

    def convert(self, origin: AcaciaExpr) -> str:
        origin_py = super().convert(origin)
        if isinstance(origin_py, int):
            return str(origin_py)
        assert isinstance(origin_py, str)
        origin_py = origin_py.lstrip()
        if origin_py.startswith("!"):
            origin_py = origin_py[1:]
        if ".." in origin_py:
            min_, max_ = origin_py.split("..")
            min_ = min_.strip()
            max_ = max_.strip()
            if not min_ and not max_:
                self.wrong_argument()
            if (min_ and not self.check_int(min_)
                or max_ and not self.check_int(max_)):
                self.wrong_argument()
            return "%s..%s" % (min_, max_)
        else:
            if self.check_int(origin_py):
                return origin_py
            else:
                self.wrong_argument()

class EFilterDataType(DefaultDataType):
    name = "Enfilter"

ctdt_efilter = CTDataType("Enfilter")

class EFilterType(Type):
    def do_init(self):
        @cmethod_of(self, "__new__")
        @axe.chop
        def _new(compiler):
            return EntityFilter()

    def datatype_hook(self):
        return EFilterDataType()

    def cdatatype_hook(self):
        return ctdt_efilter

class _EFilterData:
    def __init__(self, tag: Optional[str], subcmds: List["_ExecuteSubcmd"],
                 selector: MCSelector):
        """tag: temporary tag name (None for last tuple)"""
        self.tag = tag
        self.subcmds = subcmds
        self.selector = selector

    def copy(self):
        return _EFilterData(
            self.tag, self.subcmds.copy(), self.selector.copy()
        )

class EntityFilter(ConstExprCombined, ImmutableMixin):
    cdata_type = ctdt_efilter

    def __init__(self):
        super().__init__(EFilterDataType())
        self.context_occupied = False
        self.data: List[_EFilterData] = []
        self._new_data(compiler=None)  # initial data
        self.next_use_new_data = False
        self.cleanup: List[str] = []  # cleanup commands
        self.entity_type: Union[str, None] = None

        @cmethod_of(self, "all_players")
        @axe.chop
        @transform_immutable(self)
        def _all_players(self: EntityFilter, compiler):
            self.need_set_selector_var(compiler, "a")
            self.entity_type = PLAYER
            return self
        @cmethod_of(self, "random")
        @axe.chop
        @axe.arg("type", axe.Nullable(axe.LiteralString()), default=None,
                 rename="type_")
        @axe.arg("limit", axe.RangedLiteralInt(1, None), default=1)
        @transform_immutable(self)
        def _random(self: EntityFilter, compiler,
                    type_: Optional[str], limit: int):
            selector = self.need_set_selector_var(compiler, "r")
            if self.entity_type is None:
                if type_ is None:
                    raise axe.ArgumentError(
                        "type",
                        localize("objects.entityfilter.random.notype")
                    )
                self.entity_type = type_
            else:
                if type_ is not None and self.entity_type != type_:
                    raise axe.ArgumentError(
                        "type",
                        localize("objects.entityfilter.random.typeconflict")
                            % (type_, self.entity_type)
                    )
            if not selector.has_arg("type"):
                selector.type(type_)
            selector.limit(limit)
            # There should be a difference between these two:
            # (Supporse there are more than 5 entities with name "xxx")
            #  Enfilter().is_name("xxx").random("t", limit=5)
            #   Selects 5 random entities with name "xxx"
            #   i.e. Selects @r[type=t, c=5, name=xxx]
            #  Enfilter().random("t", limit=5).is_name("xxx")
            #   Selects at most 5 random entities with name "xxx"
            #   i.e. tag @r[type=t, c=5] add tmp
            #        Selects @e[name=xxx, tag=tmp]
            # Thus, we do not accept more arguments for @r selector
            # now (the same for nearest_from and farthest_from):
            self.next_use_new_data = True
            return self
        @cmethod_of(self, "nearest_from")
        @axe.chop
        @axe.arg("origin", PosDataType)
        @axe.arg("limit", axe.RangedLiteralInt(1, None), default=1)
        @transform_immutable(self)
        def _nearest_from(self: EntityFilter, compiler,
                          origin: "Position", limit: int):
            self.need_set_selector_var(compiler, "e")
            selector = self.need_set_context(compiler, *origin.context)
            selector.limit(limit)
            self.next_use_new_data = True  # see `_random` above
            return self
        @cmethod_of(self, "farthest_from")
        @axe.chop
        @axe.arg("origin", PosDataType)
        @axe.arg("limit", axe.RangedLiteralInt(1, None), default=1)
        @transform_immutable(self)
        def _farthest_from(self: EntityFilter, compiler,
                           origin: "Position", limit: int):
            self.need_set_selector_var(compiler, "e")
            selector = self.need_set_context(compiler, *origin.context)
            selector.limit(-limit)
            self.next_use_new_data = True  # see `_random` above
            return self
        @cmethod_of(self, "has_tag")
        @axe.chop
        @axe.star_arg("tags", axe.LiteralString())
        @transform_immutable(self)
        def _has_tag(self: EntityFilter, compiler, tags: List[str]):
            selector = self.last_selector(compiler)
            selector.tag(*tags)
            return self
        @cmethod_of(self, "has_no_tag")
        @axe.chop
        @axe.star_arg("tags", axe.LiteralString())
        @transform_immutable(self)
        def _has_no_tag(self: EntityFilter, compiler, tags: List[str]):
            selector = self.last_selector(compiler)
            selector.tag_n(*tags)
            return self
        @cmethod_of(self, "distance_from")
        @axe.chop
        @axe.arg("origin", PosDataType)
        @axe.arg("min", axe.Nullable(axe.LiteralFloat()), default=None,
                 rename="min_")
        @axe.arg("max", axe.Nullable(axe.LiteralFloat()), default=None,
                 rename="max_")
        @transform_immutable(self)
        def _distance_from(self: EntityFilter, compiler,
                           origin: "Position", min_: Optional[float],
                           max_: Optional[float]):
            selector = self.need_set_context(compiler, *origin.context)
            selector.distance(min_, max_)
            return self
        @cmethod_of(self, "is_type")
        @axe.chop
        @axe.arg("type", axe.LiteralString(), rename="type_")
        @transform_immutable(self)
        def _is_type(self: EntityFilter, compiler, type_: str):
            selector = self.new_if_got(compiler, "type")
            selector.type(type_)
            self.entity_type = type_
            return self
        @cmethod_of(self, "is_not_type")
        @axe.chop
        @axe.star_arg("types", axe.LiteralString())
        @transform_immutable(self)
        def _is_not_type(self: EntityFilter, compiler, types: List[str]):
            selector = self.last_selector(compiler)
            selector.type_n(*types)
            return self
        @cmethod_of(self, "inside")
        @axe.chop
        @axe.arg("origin", PosDataType)
        @axe.arg("dx", axe.LiteralFloat(), default=0.0)
        @axe.arg("dy", axe.LiteralFloat(), default=0.0)
        @axe.arg("dz", axe.LiteralFloat(), default=0.0)
        @transform_immutable(self)
        def _inside(self: EntityFilter, compiler,
                    origin: "Position", dx: int, dy: int, dz: int):
            selector = self.need_set_context(compiler, *origin.context)
            selector.volume(dx, dy, dz)
            return self
        @cmethod_of(self, "rot_vertical")
        @axe.chop
        @axe.arg("min", axe.LiteralFloat(), default=-90.0, rename="min_")
        @axe.arg("max", axe.LiteralFloat(), default=90.0, rename="max_")
        @transform_immutable(self)
        def _rot_vertical(self: EntityFilter, compiler,
                          min_: float, max_: float):
            selector = self.new_if_got(compiler, "rx", "rxm")
            selector.rot_vertical(min_, max_)
            return self
        @cmethod_of(self, "rot_horizontal")
        @axe.chop
        @axe.arg("min", axe.LiteralFloat(), default=-180.0, rename="min_")
        @axe.arg("max", axe.LiteralFloat(), default=180.0, rename="max_")
        @transform_immutable(self)
        def _rot_horizontal(self: EntityFilter, compiler,
                            min_: float, max_: float):
            selector = self.new_if_got(compiler, "ry", "rym")
            selector.rot_horizontal(min_, max_)
            return self
        @cmethod_of(self, "is_name")
        @axe.chop
        @axe.arg("name", axe.LiteralString())
        @transform_immutable(self)
        def _is_name(self: EntityFilter, compiler, name: str):
            selector = self.new_if_got(compiler, "name")
            selector.name(name)
            return self
        @cmethod_of(self, "is_not_name")
        @axe.chop
        @axe.star_arg("names", axe.LiteralString())
        @transform_immutable(self)
        def _is_not_name(self: EntityFilter, compiler, names: List[str]):
            selector = self.last_selector(compiler)
            selector.name_n(*names)
            return self
        @cmethod_of(self, "has_item")
        @axe.chop
        @axe.arg("item", axe.LiteralString())
        @axe.arg("quantity", IntRange(), default="1..")
        @axe.arg("data", axe.Nullable(axe.LiteralInt()), default=None)
        @axe.arg("slot_type", axe.Nullable(axe.LiteralString()), default=None)
        @axe.arg("slot_num", axe.Nullable(IntRange()), default=None)
        @transform_immutable(self)
        def _has_item(self: EntityFilter, compiler, item: str,
                      quantity: str, data: Optional[int],
                      slot_type: Optional[str], slot_num: Optional[int]):
            if slot_type is None and slot_num is not None:
                raise axe.ArgumentError(
                    "slot_num", localize("objects.entityfilter.hasitem.slot")
                )
            selector = self.last_selector(compiler)
            selector.has_item(item, quantity, data, slot_type, slot_num)
            return self
        @cmethod_of(self, "scores")
        @axe.chop
        @axe.arg("objective", axe.LiteralString())
        @axe.arg("range", IntRange(), rename="range_")
        @transform_immutable(self)
        def _scores(self: EntityFilter, compiler, objective: str,
                    range_: str):
            selector = self.last_selector(compiler)
            selector.scores(objective, range_)
            return self
        @cmethod_of(self, "level")
        @axe.chop
        @axe.arg("min", axe.Nullable(axe.LiteralInt()),
                 default=None, rename="min_")
        @axe.arg("max", axe.Nullable(axe.LiteralInt()),
                 default=None, rename="max_")
        @transform_immutable(self)
        def _level(self: EntityFilter, compiler, min_: Optional[int],
                   max_: Optional[int]):
            selector = self.new_if_got(compiler, "l", "lm")
            selector.level(min_, max_)
            self.entity_type = PLAYER
            return self
        @cmethod_of(self, "is_game_mode")
        @axe.chop
        @axe.arg("mode", axe.LiteralString())
        @transform_immutable(self)
        def _is_game_mode(self: EntityFilter, compiler, mode: str):
            selector = self.last_selector(compiler)
            selector.game_mode(mode)
            self.entity_type = PLAYER
            return self
        @cmethod_of(self, "is_not_game_mode")
        @axe.chop
        @axe.star_arg("modes", axe.LiteralString())
        @transform_immutable(self)
        def _is_not_game_mode(self: EntityFilter, compiler, modes: List[str]):
            selector = self.last_selector(compiler)
            selector.game_mode_n(*modes)
            self.entity_type = PLAYER
            return self
        @cmethod_of(self, "has_permission")
        @versionlib.only(versionlib.at_least((1, 19, 80)))
        @axe.chop
        @axe.star_arg("permissions", axe.LiteralString())
        @transform_immutable(self)
        def _has_permission(self: EntityFilter, compiler,
                            permissions: List[str]):
            selector = self.last_selector(compiler)
            selector.has_permission(*permissions)
            self.entity_type = PLAYER
            return self
        @cmethod_of(self, "has_no_permission")
        @versionlib.only(versionlib.at_least((1, 19, 80)))
        @axe.chop
        @axe.star_arg("permissions", axe.LiteralString())
        @transform_immutable(self)
        def _has_no_permission(self: EntityFilter, compiler,
                               permissions: List[str]):
            selector = self.last_selector(compiler)
            selector.has_permission_n(*permissions)
            self.entity_type = PLAYER
            return self

    def copy(self):
        res = EntityFilter()
        res.data = [data.copy() for data in self.data]
        res.context_occupied = self.context_occupied
        res.next_use_new_data = self.next_use_new_data
        res.cleanup = self.cleanup.copy()
        res.entity_type = self.entity_type
        return res

    def dump(self, command: str, among_tag: Optional[str] = None) -> CMDLIST_T:
        """
        Select entities filtered by this filter and return commands.
        The command can have "{selected}" placeholder, which will be
        replaced by the selected entity.
        When `among_tag` is specified, the filter will begin selecting
        among entities with that tag, instead of all entities.
        """
        res = []
        last_tag = among_tag
        for data in self.data[:-1]:
            selector = data.selector
            if last_tag is not None:
                selector = selector.copy()
                selector.tag(last_tag)
            res.append(cmds.Execute(
                data.subcmds,
                f"tag {selector.to_str()} add {data.tag}"
            ))
            last_tag = data.tag
        final = self.data[-1]
        final_selector = final.selector
        if last_tag is not None:
            final_selector = final_selector.copy()
            final_selector.tag(last_tag)
        res.append(cmds.Execute(
            final.subcmds, command.format(selected=final_selector.to_str())
        ))
        res.extend(self.cleanup)
        return res

    def need_set_selector_var(
        self, compiler, var: SELECTORVAR_T
    ) -> MCSelector:
        if self.data[-1].selector.is_var_set():
            self._new_data(compiler)
        res = self.data[-1].selector
        res.var = var
        return res

    def need_set_context(
        self, compiler, *context: "_ExecuteSubcmd"
    ) -> MCSelector:
        if self.context_occupied:
            self._new_data(compiler)
        self.data[-1].subcmds.extend(context)
        self.context_occupied = True
        return self.data[-1].selector

    def last_selector(self, compiler) -> MCSelector:
        if self.next_use_new_data:
            self._new_data(compiler)
        return self.data[-1].selector

    def new_if_got(self, compiler, *args: str) -> MCSelector:
        selector = self.last_selector(compiler)
        if any(selector.has_arg(a) for a in args):
            self._new_data(compiler)
        return self.data[-1].selector

    def _new_data(self, compiler: Optional["Compiler"]):
        if self.data:
            # Handle last data
            assert compiler is not None
            tag = compiler.allocate_entity_tag()
            self.data[-1].tag = tag
            self.cleanup.append("tag @e[tag={0}] remove {0}".format(tag))
        self.data.append(_EFilterData(None, [], MCSelector()))
        self.context_occupied = False
        self.next_use_new_data = False



"""
music - Generate redstone music from MIDI file
NOTE Python package `mido` is required.
"""

from typing import Dict, TYPE_CHECKING
try:
    from itertools import pairwise
except ImportError:
    def pairwise(iterable):
        """A clone of itertools.pairwise for Python 3.9 or below."""
        iterator = iter(iterable)
        last = next(iterator, None)
        for current in iterator:
            yield last, current
            last = current

from acaciamc.objects import *
from acaciamc.mccmdgen.expr import *
from acaciamc.mccmdgen.datatype import DefaultDataType
from acaciamc.mccmdgen.ctexpr import CTDataType
from acaciamc.error import *
from acaciamc.tools import axe, resultlib, cmethod_of, method_of
from acaciamc.localization import localize
import acaciamc.mccmdgen.cmds as cmds

if TYPE_CHECKING:
    from acaciamc.compiler import Compiler
    from acaciamc.mccmdgen.mcselector import MCSelector

ID2INSTRUMENT = {
    # Piano
    0: 'note.harp',
    1: 'note.harp',
    2: 'note.pling',
    3: 'note.pling',
    4: 'note.pling',
    5: 'note.pling',
    6: 'note.harp',
    7: 'note.pling',
    # Chromatic Percussion
    8: 'note.harp',
    9: 'note.bell',
    10: 'note.chime',
    11: 'note.iron_xylophone',
    12: 'note.xylophone',
    13: 'note.xylophone',
    14: 'note.chime',
    15: 'note.bell',
    # Organ
    # Organ sounds like flute somehow...
    16: 'note.flute',
    17: 'note.flute',
    18: 'note.flute',
    19: 'note.flute',
    20: 'note.flute',
    21: 'note.flute',
    22: 'note.flute',
    23: 'note.flute',
    # Guitar
    24: 'note.guitar',
    25: 'note.guitar',
    26: 'note.guitar',
    27: 'note.guitar',
    28: 'note.guitar',
    29: 'note.guitar',
    30: 'note.guitar',
    31: 'note.bass',
    # Bass
    32: 'note.bass',
    33: 'note.bass',
    34: 'note.bass',
    35: 'note.bass',
    36: 'note.bass',
    37: 'note.bass',
    38: 'note.bass',
    39: 'note.bass',
    # Solo String
    # Strings like violin sounds like flute somehow...
    40: 'note.flute',
    41: 'note.flute',
    42: 'note.flute',
    43: 'note.flute',
    44: 'note.guitar',
    45: 'note.guitar',
    46: 'note.harp',
    47: 'note.snare',
    # Ensemble
    48: 'note.flute',
    49: 'note.flute',
    50: 'note.flute',
    51: 'note.flute',
    52: 'note.flute',
    53: 'note.flute',
    # Brass
    54: 'note.flute',
    55: 'note.snare',
    56: 'note.chime',
    57: 'note.chime',
    58: 'note.chime',
    59: 'note.chime',
    60: 'note.chime',
    61: 'note.chime',
    62: 'note.chime',
    63: 'note.chime',
    # Reed
    # note.bit sounds like Sax somehow...
    64: 'note.bit',
    65: 'note.bit',
    66: 'note.bit',
    67: 'note.bit',
    68: 'note.flute',
    69: 'note.flute',
    70: 'note.flute',
    71: 'note.flute',
    # Pipe
    72: 'note.flute',
    73: 'note.flute',
    74: 'note.flute',
    75: 'note.flute',
    76: 'note.flute',
    77: 'note.flute',
    78: 'note.bell',
    79: 'note.flute',
    # Synth Lead
    80: 'note.bit',
    81: 'note.bit',
    82: 'note.flute',
    83: 'note.flute',
    84: 'note.guitar',
    85: 'note.bit',
    86: 'note.bit',
    87: 'note.bit',
    # Synth Pad
    88: 'note.bit',
    89: 'note.bit',
    90: 'note.bit',
    91: 'note.bit',
    92: 'note.guitar',
    93: 'note.bit',
    94: 'note.bit',
    95: 'note.guitar',
    # Synth Effect
    96: 'note.bit',
    97: 'note.bit',
    98: 'note.bit',
    99: 'note.bit',
    100: 'note.bit',
    101: 'note.bit',
    102: 'note.bit',
    103: 'note.bit',
    # Ethnic
    104: 'note.guitar',
    105: 'note.banjo',
    106: 'note.guitar',
    107: 'note.guitar',
    108: 'note.bell',
    109: 'note.flute',
    110: 'note.guitar',
    111: 'note.flute',
    # Percussive
    112: 'note.bell',
    113: 'note.bell',
    114: 'note.drum',
    115: 'note.cow_bell',
    116: 'note.drum',
    117: 'note.drum',
    118: 'note.drum',
    119: 'note.bit',
    # Sound Effects
    120: 'note.hat',
    121: 'note.hat',
    122: 'note.hat',
    123: 'note.hat',
    124: 'note.hat',
    125: 'note.hat',
    126: 'note.hat',
    127: 'note.snare'
}

class MusicDataType(DefaultDataType):
    name = "Music"

ctdt_music = CTDataType("music")

class MusicType(Type):
    """
    Music(
        path: str,
        looping: bool-literal = False,
        loop_interval: int-literal = 50,
        listener: PlayerSelector = <all players>,
        note_offset: int-literal = 0,
        chunk_size: int-literal = 500,
        speed: float = 1.0,
        volume: float = 1.0,
        channel_volume: map[int-literal, float] = {:},
        instrument: map[int-literal, str] = {:},
    )

    A music to generate from MIDI file `path`.
    If `looping` is True, the music will be played repeatingly, with
    interval `loop_interval` before replaying.
    `listener` specifies who would hear the music.
    `note_offset` would make the whole music higher if it is positive,
    or lower if negative.
    Since music runs a LOT of commands every tick, we seperate commands
    into several files to make sure you are not running too many commands
    every tick. `chunk_size` determines how many commands there will be
    in 1 file.
    `speed` sets the speed of music.
    `volume` sets the overall volume of music.
    `channel_volume` sets the volume factor of each MIDI channel (0-15).
    By default, all channels have the same volume factor (1.0).
    `instrument` sets the corresponding Minecraft sound of each MIDI
    instrument (0-127). The default mapping is in `ID2INSTRUMENT`. An
    example: {127: "note.hat"}.
    """
    def do_init(self):
        @cmethod_of(self, "__new__")
        @axe.chop
        @axe.arg("path", axe.LiteralString())
        @axe.arg("looping", axe.LiteralBool(), default=False)
        @axe.arg("loop_interval", axe.LiteralInt(), default=50)
        @axe.arg("listener", axe.PlayerSelector(), default=None)
        @axe.arg("note_offset", axe.LiteralInt(), default=0)
        @axe.arg("chunk_size", axe.RangedLiteralInt(1, None), default=500)
        @axe.arg("speed", axe.LiteralFloat(), default=1.0)
        @axe.arg("volume", axe.LiteralFloat(), default=1.0)
        @axe.arg("channel_volume", axe.MapOf(
            axe.RangedLiteralInt(0, 15), axe.LiteralFloat()
        ), default={})
        @axe.arg("instrument", axe.MapOf(
            axe.RangedLiteralInt(0, 127), axe.LiteralString()
        ), default={})
        def _new(compiler, path: str, looping: bool, loop_interval: int,
                 listener: "MCSelector", note_offset: int,
                 chunk_size: int, speed: float, volume: float,
                 channel_volume: Dict[int, float], instrument: Dict[int, str]):
            try:
                midi = mido.MidiFile(path)
            except OSError as err:
                raise Error(
                    ErrorType.IO,
                    message=localize("modules.music.doinit.midiparser")
                        % err.strerror
                )
            if speed <= 0:
                raise axe.ArgumentError(
                    "speed", localize("modules.music.doinit.mustpos")
                )
            if volume <= 0:
                raise axe.ArgumentError(
                    "volume", localize("modules.music.doinit.mustpos")
                )
            if any([v < 0 for v in channel_volume.values()]):
                raise axe.ArgumentError(
                    "channel_volume", localize("modules.music.doinit.mustpos")
                )
            looping_info = loop_interval if looping else -1
            if listener is None:
                listener_str = "@a"
            else:
                listener_str = listener.to_str()
            return Music(
                midi, listener_str, looping_info, note_offset,
                chunk_size, speed, volume, channel_volume, instrument,
                compiler
            )

    def datatype_hook(self):
        return MusicDataType()

    def cdatatype_hook(self):
        return ctdt_music

class Music(ConstExprCombined):
    # NOTE We are using `MT` to refer to 1 MIDI tick and `GT` for 1 MC game
    # tick.

    cdata_type = ctdt_music

    def __init__(self, midi, listener_str: str, looping: int,
                 note_offset: int, chunk_size: int, speed: float,
                 volume: float, channel_volume: Dict[int, float],
                 instrument: Dict[int, str], compiler: "Compiler"):
        super().__init__(MusicDataType())
        self.midi = midi
        self.listener_str = listener_str
        self.note_offset = note_offset
        self.chunk_size = chunk_size
        self.override_instrument = instrument
        self.tracks = [t.copy() for t in midi.tracks]
        # Check MIDI type
        if midi.type != 0 and midi.type != 1:
            raise Error(
                ErrorType.ANY,
                message=localize("modules.music.music.init.unsupported")
                    % midi.type
            )
        # Speed settings
        self.bpm = 120
        self.mt_per_beat = midi.ticks_per_beat
        self.user_speed = speed
        # Ticking
        self.mt = 0
        self.gt = 0.0
        self.gt_int = 0  # Always == round(self.gt)
        self.last_gt_int = 0
        # last_msg_mt: track id to MT when last Message is handled
        self.last_msg_mt = dict.fromkeys(range(len(midi.tracks)), 0)
        # Channel info
        self.channel_volume = {}  # channel id to volume (0-15)
        self.channel_instrument = {}  # channel id to instrument id
        ## Default instrument: 0~8 & 10~15: Piano (0); 9: Drum set (127)
        ## Default volume: 100
        for i in range(16):
            self.channel_instrument[i] = 0
            self.channel_volume[i] = 100
        self.channel_instrument[9] = 127
        # Timer:
        #   when 0 <= timer <= music length, the music is playing
        #   when timer > music length, the music has ended
        #   only when looping is enabled, if timer == music length + 1,
        #       it will be reset next GT so that it will play again
        #   when timer < 0, it's the countdown before we start playing
        self.timer = IntVar.new(compiler)
        # Volume
        self.user_volume = volume
        self.user_channel_volume = dict.fromkeys(range(16), 1.0)
        self.user_channel_volume.update(channel_volume)
        # Create file
        self.files = []
        # file_sep_gt: in which GT we seperate the file
        self.file_sep_gt = []
        self.cur_chunk_size = 0  # Commands written in current file
        self.new_file()  # Initial file
        # Go
        while not self.is_finished():
            self.main_loop()
        GT_LEN = self.gt_int  # Length of music in GT
        # The last file may be useless
        if not self.files[-1].has_content():
            self.files.pop()
            self.file_sep_gt.pop()
        # Add file
        for file in self.files:
            compiler.add_file(file)
        self.file_sep_gt.append(GT_LEN + 1)
        # Loop commands
        loopcmds: CMDLIST_T = [cmds.Comment("# music.Music")]
        loopcmds.extend(
            cmds.Execute(
                [cmds.ExecuteScoreMatch(self.timer.slot, f"{t1}..{t2 - 1}")],
                runs=cmds.InvokeFunction(self.files[i])
            )
            for i, (t1, t2) in enumerate(pairwise(self.file_sep_gt))
        )
        loopcmds.append(
            cmds.Execute(
                [cmds.ExecuteScoreMatch(self.timer.slot, f"..{GT_LEN}")],
                runs=cmds.ScbAddConst(self.timer.slot, 1)
            )
        )
        if looping >= 0:
            loopcmds.append(
                cmds.Execute(
                    [cmds.ExecuteScoreMatch(self.timer.slot, str(GT_LEN + 1))],
                    runs=cmds.ScbSetConst(self.timer.slot, -looping)
                )
            )
        # Register loop commands to be called every tick
        compiler.file_tick.extend(loopcmds)
        # Create attributes
        self.attribute_table.set("_timer", self.timer)
        self.attribute_table.set("LENGTH", IntLiteral(GT_LEN))
        @method_of(self, "play")
        @axe.chop
        @axe.arg("timer", IntDataType, default=IntLiteral(0))
        def _play(compiler, timer: AcaciaExpr):
            """
            .play(timer: int = 0)

            Start playing the music. When `timer` < 0, its the delay
            of playing. When `timer` >= 0, its where the music starts
            playing.
            """
            commands = timer.export(self.timer, compiler)
            return resultlib.commands(commands)
        @method_of(self, "stop")
        @axe.chop
        def _stop(compiler):
            """.stop(): Stop the music"""
            commands = [cmds.ScbSetConst(self.timer.slot, GT_LEN + 2)]
            return resultlib.commands(commands)

    def main_loop(self):
        # Read messages
        for i, track in enumerate(self.tracks):
            if not track:
                continue
            message = track[0]
            if message.time > self.mt - self.last_msg_mt[i]:
                continue
            # Handle message
            mtype = message.type
            if mtype == "note_on":
                if message.velocity != 0:
                    self.play_note(message)
            elif mtype == "set_tempo":
                self.bpm = 6E+7 / message.tempo
            elif message.is_cc():
                if message.control == 7:  # Volume
                    self.channel_volume[message.channel] = message.value
            elif mtype == "program_change":
                self.channel_instrument[message.channel] = message.program
            # Reset timer
            self.last_msg_mt[i] = self.mt
            track.pop(0)
        # Time increment
        self.mt += 1
        # bpm * mt_per_beat is MT per minute. Divide it by 1200 to get MT
        # per GT, and multiply it by `user_speed` at last
        self.gt += 1 / (self.bpm * self.mt_per_beat * self.user_speed / 1200)
        self.gt_int = round(self.gt)
        # Update last_gt_int
        if self.gt_int > self.last_gt_int:
            self.last_gt_int = self.gt_int
            # New file check
            if self.cur_chunk_size >= self.chunk_size:
                self.new_file()

    def new_file(self):
        self.cur_file = cmds.MCFunctionFile()
        self.files.append(self.cur_file)
        self.cur_file.write_debug("# Music loop")
        self.file_sep_gt.append(self.gt_int)
        self.cur_chunk_size = 0

    def is_finished(self):
        return not any(self.tracks)

    def get_instrument(self, channel: int) -> str:
        """Get MC sound of channel."""
        ins_id = self.channel_instrument[channel]
        return self.override_instrument.get(ins_id, ID2INSTRUMENT[ins_id])

    def get_volume(self, channel: int, velocity: int) -> float:
        """Get MC volume (0~1) according to channel and velocity."""
        channel_v = (self.channel_volume[channel]
                     * self.user_channel_volume[channel])
        return velocity * channel_v / 127 / 127 * self.user_volume

    def get_pitch(self, note: int) -> float:
        """Get MC pitch from MIDI note"""
        return 2 ** ((note + self.note_offset - 54) / 12 - 1)

    def play_note(self, message):
        """Play a note according to note_on Message"""
        volume = self.get_volume(message.channel, message.velocity)
        if volume == 0:
            return
        pitch = self.get_pitch(message.note)
        sound = self.get_instrument(message.channel)
        self.cur_file.write(cmds.Execute(
            [cmds.ExecuteScoreMatch(self.timer.slot, str(self.gt_int)),
             cmds.ExecuteEnv("as", self.listener_str),
             cmds.ExecuteEnv("at", "@s")],
            runs=cmds.Cmd(
                "playsound %s @s ~ ~ ~ %.2f %.3f" % (sound, volume, pitch)
            )
        ))
        self.cur_chunk_size += 1

def acacia_build(compiler: "Compiler"):
    global mido
    try:
        import mido
    except ImportError:
        raise Error(
            ErrorType.ANY,
            message=localize("modules.music.acaciabuild.norequire")
        )
    return {"Music": MusicType()}



"""math - Math related utilities."""
from typing import List, Optional, TYPE_CHECKING

from acaciamc.objects import *
from acaciamc.mccmdgen.expr import *
from acaciamc.objects.integer import (
    IntRandom, IntCmdOp, IntOpSelf, IntOpVar, IntOp, STR2SCBOP
)
from acaciamc.ast import ModuleMeta
from acaciamc.tools import axe
from acaciamc.constants import INT_MIN, INT_MAX
from acaciamc.localization import localize

if TYPE_CHECKING:
    from acaciamc.compiler import Compiler

def internal(name: str) -> AcaciaExpr:
    res = _math.attribute_table.lookup(name)
    if res is None:
        raise ValueError(name)
    return res

@axe.chop
@axe.arg("min", axe.LiteralInt(), rename="min_")
@axe.arg("max", axe.LiteralInt(), rename="max_")
def _randint(compiler, min_: int, max_: int):
    """
    randint(const min: int, const max: int) -> int
    Get a random integer between `min` and `max` (inclusive).
    """
    return IntOpGroup(init=IntRandom(min_, max_))

class IntXOpSelf(IntOp):
    def __init__(self, x: cmds.ScbSlot, op: str) -> None:
        self.op = op
        self.x = x

    def scb_did_assign(self, slot: cmds.ScbSlot) -> bool:
        return slot == self.x

    def resolve(self, var: IntVar) -> CMDLIST_T:
        return [cmds.ScbOperation(STR2SCBOP[self.op], self.x, var.slot)]

@axe.chop
@axe.arg("base", IntDataType)
@axe.arg("exp", IntDataType)
def _pow(compiler: "Compiler", base, exp):
    """
    pow(base: int, exp: int) -> int
    return "base" to the power of "exp".
    `pow(0, 0)` is defined to be 1.
    Negative values of `exp` result in undefined behavior.
    """
    if not isinstance(exp, IntLiteral):
        # Fallback to `_math.pow`
        return internal("pow").call([base, exp], {}, compiler)
    if exp.value < 0:
        raise axe.ArgumentError('exp', localize("modules.math.pow.negative"))
    if exp.value == 0:
        return IntLiteral(1)
    # Optimize when base is a literal
    if isinstance(base, IntLiteral):
        return IntLiteral(base.value ** exp.value)
    # Base is a variable but exp is a literal
    # Reference: https://en.wikipedia.org/wiki/Exponentiation_by_squaring
    # Section "With constant auxiliary memory"
    res = IntOpGroup.from_intexpr(base)
    e = exp.value
    y: Optional[cmds.ScbSlot] = None
    while e != 1:
        if e % 2:  # x & 1 is in fact slower than x % 2 in Python
            if y is None:
                y = compiler.allocate_tmp()
                res.add_op(IntXOpSelf(y, "="))
            else:
                res.add_op(IntXOpSelf(y, "*"))
        res.add_op(IntOpSelf("*"))
        e //= 2  # x >>= 1 is slower than x //= 2, too
    if y is not None:
        res.add_op(IntOpVar("*", y))
    return res

class IntRestrict(IntOp):
    def __init__(self, value: int, is_upper: bool):
        self.value = value
        self.is_upper = is_upper

    def resolve(self, var: IntVar):
        if ((self.is_upper and self.value == INT_MAX)
                or (not self.is_upper and self.value == INT_MIN)):
            return []
        rng = ((str(self.value + 1) + "..") if self.is_upper
               else ".." + str(self.value - 1))
        return [
            cmds.Execute(
                [cmds.ExecuteScoreMatch(var.slot, rng)],
                cmds.ScbSetConst(var.slot, self.value)
            )
        ]

@axe.chop
@axe.star_arg("operands", IntDataType)
def _min(compiler, operands: List[AcaciaExpr]):
    """
    min(*operands: int) -> int
    Return the minimum value among `args`.
    """
    if not operands:
        raise axe.ArgumentError('operands', localize("modules.math.min.empty"))
    # Get literals
    upper = None
    rest = []
    for operand in operands:
        if isinstance(operand, IntLiteral):
            if upper is None or operand.value < upper:
                upper = operand.value
        else:
            rest.append(operand)
    if not rest:
        return IntLiteral(upper)
    # Get first arg
    res = IntOpGroup.from_intexpr(rest[0])
    # Handle args left
    for operand in rest[1:]:
        deps, var = to_IntVar(operand, compiler)
        res.add_op(IntCmdOp(deps))
        res.add_op(IntOpVar("<", var.slot))
    if upper is not None:
        res.add_op(IntRestrict(upper, is_upper=True))
    return res

@axe.chop
@axe.star_arg("operands", IntDataType)
def _max(compiler, operands: List[AcaciaExpr]):
    """
    max(*operands: int) -> int
    Return the maximum value among `args`.
    """
    if not operands:
        raise axe.ArgumentError('operands', localize("modules.math.max.empty"))
    # Get literals
    lower = None
    rest = []
    for operand in operands:
        if isinstance(operand, IntLiteral):
            if lower is None or operand.value > lower:
                lower = operand.value
        else:
            rest.append(operand)
    if not rest:
        return IntLiteral(lower)
    # Get first arg
    res = IntOpGroup.from_intexpr(rest[0])
    # Handle args left
    for operand in rest[1:]:
        deps, var = to_IntVar(operand, compiler)
        res.add_op(IntCmdOp(deps))
        res.add_op(IntOpVar(">", var.slot))
    if lower is not None:
        res.add_op(IntRestrict(lower, is_upper=False))
    return res

@axe.chop
@axe.arg("x", IntDataType)
@axe.arg("y", IntDataType)
def _mod(compiler, x, y):
    if isinstance(x, IntLiteral) and isinstance(y, IntLiteral):
        if y.value == 0:
            raise axe.ArgumentError('y', localize("modules.math.mod.zero"))
        return IntLiteral(x.value % y.value)
    return internal("mod").call([x, y], {}, compiler)

@axe.chop
@axe.arg("x", IntDataType)
@axe.arg("y", IntDataType)
def _floordiv(compiler, x, y):
    if isinstance(x, IntLiteral) and isinstance(y, IntLiteral):
        if y.value == 0:
            raise axe.ArgumentError(
                'y', localize("modules.math.floordiv.zero")
            )
        return IntLiteral(x.value // y.value)
    return internal("floordiv").call([x, y], {}, compiler)

def acacia_build(compiler: "Compiler"):
    global _math
    f = cmds.MCFunctionFile()
    _math = compiler.get_module(ModuleMeta("_math"), f)
    attributes = {
        'randint': BinaryFunction(_randint),
        'pow': BinaryCTFunction(_pow),
        'min': BinaryCTFunction(_min),
        'max': BinaryCTFunction(_max),
        'mod': BinaryCTFunction(_mod),
        'floordiv': BinaryCTFunction(_floordiv),
    }
    return BuiltModule(attributes, f.commands)



"""Miscellaneous utility stuffs."""

__all__ = ["InvalidOpError", "unreachable", "apply_decorators"]

from typing import Reversible

class InvalidOpError(Exception):
    """Can be raised in some methods to indicate an invalid operation."""
    pass

def unreachable(message: str = ""):
    """Indicates an unreachable code path."""
    suffix = f": {message}" if message else ""
    raise AssertionError("unreachable() called" + suffix)

def apply_decorators(decorators: Reversible):
    """
    Return a decorator that applies a sequence of decorators.
        @apply_decorators([a, b])
    is equivalent to
        @a
        @b
    """
    def decorator(func):
        for deco in reversed(decorators):
            func = deco(func)
        return func
    return decorator
